"""
Thread-safe file downloading and cacheing

Authors
  * Leo 2022
  * Cheng Liang 2022
"""

import hashlib
import logging
import os
import shutil
import sys
import tempfile
import time
from pathlib import Path
from urllib.request import Request, urlopen

import requests
from filelock import FileLock
from tqdm import tqdm

logger = logging.getLogger(__name__)


_download_dir = Path.home() / ".cache" / "s3prl" / "download"

__all__ = [
    "get_dir",
    "set_dir",
    "download",
    "urls_to_filepaths",
]


def get_dir():
    _download_dir.mkdir(exist_ok=True, parents=True)
    return _download_dir


def set_dir(d):
    global _download_dir
    _download_dir = Path(d)


def _download_url_to_file(url, dst, hash_prefix=None, progress=True):
    """
    This function is not thread-safe. Please ensure only a single
    thread or process can enter this block at the same time
    """

    file_size = None
    req = Request(url, headers={"User-Agent": "torch.hub"})
    u = urlopen(req)
    meta = u.info()
    if hasattr(meta, "getheaders"):
        content_length = meta.getheaders("Content-Length")
    else:
        content_length = meta.get_all("Content-Length")
    if content_length is not None and len(content_length) > 0:
        file_size = int(content_length[0])

    dst = os.path.expanduser(dst)
    dst_dir = os.path.dirname(dst)
    f = tempfile.NamedTemporaryFile(delete=False, dir=dst_dir)

    try:
        sha256 = None
        if hash_prefix is not None:
            sha256 = hashlib.sha256()

        tqdm.write(f"Downloading: {url}", file=sys.stderr)
        tqdm.write(f"Destination: {dst}", file=sys.stderr)
        with tqdm(
            total=file_size,
            disable=not progress,
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
        ) as pbar:
            while True:
                buffer = u.read(8192)
                if len(buffer) == 0:
                    break
                f.write(buffer)
                if hash_prefix is not None and sha256 is not None:
                    sha256.update(buffer)
                pbar.update(len(buffer))

        f.close()
        if hash_prefix is not None and sha256 is not None:
            digest = sha256.hexdigest()
            if digest[: len(hash_prefix)] != hash_prefix:
                raise RuntimeError(
                    'invalid hash value (expected "{}", got "{}")'.format(
                        hash_prefix, digest
                    )
                )
        shutil.move(f.name, dst)
    finally:
        f.close()
        if os.path.exists(f.name):
            os.remove(f.name)


def _download_url_to_file_requests(url, dst, hash_prefix=None, progress=True):
    """
    Alternative download when urllib.Request fails.
    """

    req = requests.get(url, stream=True, headers={"User-Agent": "torch.hub"})
    file_size = int(req.headers["Content-Length"])

    dst = os.path.expanduser(dst)
    dst_dir = os.path.dirname(dst)
    f = tempfile.NamedTemporaryFile(delete=False, dir=dst_dir)

    try:
        sha256 = None
        if hash_prefix is not None:
            sha256 = hashlib.sha256()

        tqdm.write(
            f"urllib.Request method failed. Trying using another method...",
            file=sys.stderr,
        )
        tqdm.write(f"Downloading: {url}", file=sys.stderr)
        tqdm.write(f"Destination: {dst}", file=sys.stderr)
        with tqdm(
            total=file_size,
            disable=not progress,
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
        ) as pbar:
            for chunk in req.iter_content(chunk_size=1024 * 1024 * 10):
                if chunk:
                    f.write(chunk)
                    f.flush()
                    os.fsync(f.fileno())
                    if hash_prefix is not None and sha256 is not None:
                        sha256.update(chunk)
                    pbar.update(len(chunk))

        f.close()
        if hash_prefix is not None and sha256 is not None:
            digest = sha256.hexdigest()
            if digest[: len(hash_prefix)] != hash_prefix:
                raise RuntimeError(
                    'invalid hash value (expected "{}", got "{}")'.format(
                        hash_prefix, digest
                    )
                )
        shutil.move(f.name, dst)
    finally:
        f.close()
        if os.path.exists(f.name):
            os.remove(f.name)


def _download(filepath: Path, url, refresh: bool, new_enough_secs: float = 2.0):
    """
    If refresh is True, check the latest modfieid time of the filepath.
    If the file is new enough (no older than `new_enough_secs`), than directly use it.
    If the file is older than `new_enough_secs`, than re-download the file.
    This function is useful when multi-processes are all downloading the same large file
    """

    Path(filepath).parent.mkdir(exist_ok=True, parents=True)

    lock_file = Path(str(filepath) + ".lock")
    logger.info(f"Requesting URL: {url}")

    with FileLock(str(lock_file)):
        if not filepath.is_file() or (
            refresh and (time.time() - os.path.getmtime(filepath)) > new_enough_secs
        ):
            try:
                _download_url_to_file(url, filepath)
            except:
                _download_url_to_file_requests(url, filepath)

    logger.info(f"Using URL's local file: {filepath}")


def _urls_to_filepaths(*args, refresh=False, download: bool = True):
    """
    Preprocess the URL specified in *args into local file paths after downloading

    Args:
        Any number of URLs (1 ~ any)

    Return:
        Same number of downloaded file paths
    """

    def _url_to_filepath(url):
        assert isinstance(url, str)
        m = hashlib.sha256()
        m.update(str.encode(url))
        filepath = get_dir() / f"{str(m.hexdigest())}.{Path(url).name}"
        if download:
            _download(filepath, url, refresh=refresh)
        return str(filepath.resolve())

    paths = [_url_to_filepath(url) for url in args]
    return paths if len(paths) > 1 else paths[0]


download = _download
urls_to_filepaths = _urls_to_filepaths



"""
Define how a model is trained & evaluated for each step in the train/valid/test loop
"""

from .base import Task
from .diarization import DiarizationPIT
from .dump_feature import DumpFeature
from .speaker_verification_task import SpeakerVerification
from .speech2text_ctc_task import Speech2TextCTCTask
from .utterance_classification_task import (
    UtteranceClassificationTask,
    UtteranceMultiClassClassificationTask,
)

__all__ = [
    "Task",
    "DiarizationPIT",
    "DumpFeature",
    "SpeakerVerification",
    "Speech2TextCTCTask",
    "UtteranceClassificationTask",
    "UtteranceMultiClassClassificationTask",
]



"""
Create vocabulary (train tokenizer)

Authors:
  * Heng-Jui Chang 2022
"""

import logging
import os
import tempfile
from collections import Counter
from typing import List, Union

logger = logging.getLogger(__name__)

__all__ = ["generate_basic_vocab", "generate_subword_vocab", "generate_vocab"]


def generate_basic_vocab(
    mode: str,
    text_list: List[str],
    vocab_size: int = -1,
    coverage: float = 1.0,
    sort_vocab: bool = True,
) -> List[str]:
    """Generates basic vocabularies, including character and word-based vocabularies.

    Args:
        mode (str): Vocabulary type (character or word).
        text_list (List[str]): List of text data.
        vocab_size (int, optional):
            Vocabulary size, if not specified, vocab_size would be `coverage * actual vocab size`. Defaults to -1.
        coverage (float, optional): Vocabulary coverage. Defaults to 1.0.
        sort_vocab (bool, optional): Sort vocabularies alphabetically. Defaults to True.

    Returns:
        List[str]: A list of vocabularies.
    """

    assert mode in {"character", "word"}, mode
    assert vocab_size == -1 or vocab_size > 0, vocab_size
    assert coverage > 0.0 and coverage <= 1.0, coverage

    logger.info(
        f"Generating vocab (type = {mode}, coverage = {coverage}) from {len(text_list)} sentences."
    )

    counter = Counter()

    for text in text_list:
        if mode == "character":
            counter.update(text)
        if mode == "word":
            counter.update(text.split())

    if vocab_size < 0:
        vocab_size = int(len(counter) * coverage)
    else:
        vocab_size = min(vocab_size, len(counter))

    if vocab_size < len(counter):
        vocab_list = sorted(counter.keys(), key=lambda k: counter[k], reverse=True)
        vocab_list = vocab_list[:vocab_size]
    else:
        vocab_list = list(counter.keys())

    if sort_vocab:
        vocab_list = sorted(vocab_list)

    logger.info(f"Generated {vocab_size} {mode} vocabularies.")

    return vocab_list


def generate_subword_vocab(
    text_list: List[str] = None,
    text_file: str = None,
    output_file: str = None,
    vocab_size: int = 1000,
    character_coverage: float = 1.0,
) -> str:
    """Generates subword vocabularies based on `sentencepiece`.

    Args:
        text_list (List[str], optional): List of text data. Defaults to None.
        text_file (str, optional): Path to text data. Defaults to None.
        output_file (str, optional): Path to save trained subword vocabularies. Defaults to "".
        vocab_size (int, optional): Vocabulary size. Defaults to 8000.
        character_coverage (float, optional): Coverage of characters in text data. Defaults to 1.0.

    Raises:
        ImportError: If `sentencepiece` is not installed.

    Returns:
        str: Path to `${output_file}.model`.
    """

    try:
        import sentencepiece as splib
    except ImportError:
        raise ImportError(
            "`sentencepiece` cannot be imported, please run `pip install sentencepiece` first"
        )

    assert output_file is not None
    output_file = str(output_file)
    assert vocab_size > 0, vocab_size

    cmd = (
        "--input={} --model_prefix={} --model_type=unigram "
        "--vocab_size={} --character_coverage={} "
        "--pad_id=0 --eos_id=1 --unk_id=2 --bos_id=-1 "
        "--eos_piece=<eos> --remove_extra_whitespaces=true "
    )

    if text_list is not None:
        assert isinstance(text_list, list)
        assert isinstance(text_list[0], str)

        logger.info(
            f"Generating vocab (type = subword, coverage = {character_coverage}) from {len(text_list)} sentences."
        )

        with tempfile.TemporaryDirectory() as directory:
            input_file = os.path.join(directory, "text.txt")
            with open(input_file, "w") as fp:
                for text in text_list:
                    fp.write(text + "\n")

            cmd = cmd.format(
                input_file,
                output_file,
                vocab_size,
                character_coverage,
            )
            splib.SentencePieceTrainer.Train(cmd)

    if text_file is not None:
        logger.info(
            f"Generating vocab (type = subword, coverage = {character_coverage}) from {text_file}"
        )

        cmd = cmd.format(
            text_file,
            output_file,
            vocab_size,
            character_coverage,
        )
        splib.SentencePieceTrainer.Train(cmd)

    return output_file + ".model"


def generate_vocab(
    mode: str,
    text_list: List[str] = None,
    text_file: str = None,
    read_lines: int = 10000000,
    **vocab_args,
) -> Union[List[str], str]:
    """Generates vocabularies given text data.

    Args:
        mode (str): Vocabulary type
        text_list (List[str], optional): List of text data. Defaults to None.
        text_file (str, optional): Path to text data. Defaults to None.
        read_lines (int, optional): Maximum lines to read from `text_file`. Defaults to 10000000.
        vocab_args:
            if :code:`mode != subword`, arguments for :obj:`generate_basic_vocab`
            if :code:`mode == subword`, arguments for :obj:`generate_subword_vocab`

    Returns:
        Union[List[str], str]: A list of vocabularies or a path to `.vocab` file.
    """

    if text_list is None and mode in {"character", "word", "phoneme"}:
        assert isinstance(text_file, str)
        with open(text_file, "r", encoding="UTF-8") as fp:
            text_list = [
                line.strip("\r\n ") for i, line in enumerate(fp) if i < read_lines
            ]

    if mode == "character":
        return generate_basic_vocab("character", text_list, **vocab_args)
    if mode in {"word", "phoneme"}:
        return generate_basic_vocab("word", text_list, **vocab_args)
    if mode == "subword":
        return generate_subword_vocab(
            text_list=text_list, text_file=text_file, **vocab_args
        )
    else:
        raise ValueError(f"Unsupported mode (vocabulary type): {mode}")



# Copyright (c) 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in
# the root directory of this source tree. An additional grant of patent rights
# can be found in the PATENTS file in the same directory.

import logging
import os
import sys
from dataclasses import dataclass, field
from typing import Dict, List, Optional, Tuple

import numpy as np
from omegaconf import MISSING

from ..wav2vec2.wav2vec2_model import ChoiceEnum

logger = logging.getLogger(__name__)

MASK_TYPE_CHOICES = ChoiceEnum(
    ["retain_spans", "random_mask", "random_mask_batched", "chunk_mask"]
)


@dataclass
class MAE_AST_Pretraining_Config:
    data: str = field(default=MISSING, metadata={"help": "path to data directory"})

    sample_rate: int = field(
        default=16_000,
        metadata={
            "help": "target sample rate. audio files will be up/down "
            "sampled to this rate"
        },
    )
    normalize: bool = field(
        default=False,
        metadata={"help": "if set, normalizes input to have 0 mean and unit variance"},
    )
    enable_padding: bool = field(
        default=False,
        metadata={"help": "pad shorter samples instead of cropping"},
    )
    max_keep_size: Optional[int] = field(
        default=None,
        metadata={"help": "exclude sample longer than this"},
    )
    max_sample_size: Optional[int] = field(
        default=None,
        metadata={"help": "max sample size to crop to for batching"},
    )
    min_sample_size: Optional[int] = field(
        default=None,
        metadata={"help": "min sample size to crop to for batching"},
    )
    random_crop: Optional[bool] = field(
        default=True,
        metadata={"help": "always crop from the beginning if false"},
    )
    pad_audio: Optional[bool] = field(
        default=False,
        metadata={"help": "pad audio to the longest one in the batch if true"},
    )

    feature_type: Optional[str] = field(
        default="wav",
        metadata={"help": "choose from ['wav', 'spectrogram', 'fbank', 'mfcc']"},
    )

    feature_rate: Optional[int] = field(
        default=100,
        metadata={
            "help": "rate of feature input to the transformer, if use wav, this arg is omited, else if use spectrogram/fbank/mfcc, the default is 100, i.e. 1s audio gives 100 frames. the label rate of using MFCC is also 100"
        },
    )

    feature_dim: Optional[int] = field(
        default=100,
        metadata={
            "help": "dim feature input to the transformer, if use wav, this arg is omited, else if use spectrogram/fbank/mfcc, the default is 80"
        },
    )

    deltas: Optional[bool] = field(
        default=True,
        metadata={
            "help": "whether or not add delta and delta-delta to the feature, only effective for spectrogram/fbank/mfcc"
        },
    )

    mask_spans: Optional[bool] = field(
        default=False,
        metadata={"help": "mask random spans, same as that is used in HuBERT and w2v2"},
    )

    mask_type: MASK_TYPE_CHOICES = field(
        default="random_mask",
        metadata={
            "help": """Determine type of mask for MAE pretraining. 
                      -retain_spans: Only for frame data. Wav2Vec2 like masking.
                      -random_mask: Perform masking on completely random tokens. No chunking. Used in MAE
                      -random_mask_batched: random_mask with the same mask across the batch.
                      -chunk_mask: Perform masking on chunks until mask_spans hit. From SSAST. Same across batch for speed.
                          """
        },
    )


class MAE_AST_Pretraining_Task:
    cfg: MAE_AST_Pretraining_Config

    def __init__(
        self,
        cfg: MAE_AST_Pretraining_Config,
    ) -> None:
        super().__init__(cfg)

        logger.info(f"current directory is {os.getcwd()}")
        logger.info(f"MAEPretrainingTask Config {cfg}")

        self.cfg = cfg

    @property
    def source_dictionary(self):
        return None

    @property
    def target_dictionary(self):
        return None

    @property
    def dictionaries(self):
        return None

    @classmethod
    def setup_task(
        cls, cfg: MAE_AST_Pretraining_Config, **kwargs
    ) -> "MAE_AST_Pretraining_Task":
        return cls(cfg)

    def load_dataset(self, split: str, **kwargs) -> None:
        raise NotImplementedError(
            "This part is not used thus not included in S3PRL, please visit the official repo."
        )

    def max_positions(self) -> Tuple[int, int]:
        return (sys.maxsize, sys.maxsize)

    def filter_indices_by_size(self, indices: np.array, *args, **kwargs) -> np.array:
        return indices



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
# 	Alan Baade originally adopted from Fast VGS Wrapper by Puyuan Peng,
#   later modified by the S3PRL team.
"""*********************************************************************************************"""
#   FileName     [ upstream/mae_ast/expert.py ]
#   Synopsis     [ Upstream MAE-AST Wrapper ]
#   Author       [ Alan Baade, S3PRL ]
#   Copyright    [ Copyleft(c), Alan Baade ]
"""*********************************************************************************************"""

from types import SimpleNamespace

import torch
import torchaudio
from torch.nn.utils.rnn import pad_sequence

from ..interfaces import UpstreamBase
from .mae_ast import MAE_AST

FBANK_SAMPLE_STRIDE = 160


class UpstreamExpert(UpstreamBase):
    def __init__(self, ckpt, **kwargs):
        super().__init__(**kwargs)

        checkpoint = torch.load(ckpt, map_location="cpu")

        self.cfg = checkpoint["cfg"]["model"]
        self.task_cfg = checkpoint["cfg"]["task"]

        self.model = MAE_AST(
            SimpleNamespace(**checkpoint["cfg"]["model"]),
            SimpleNamespace(**checkpoint["cfg"]["task"]),
        ).cpu()

        self.model.load_state_dict(checkpoint["model"], strict=True)

        # Required for hidden states to have defined indices.
        self.model.encoder.layerdrop = 0

        self.sample_rate = self.task_cfg["sample_rate"]
        self.feature_dim = self.task_cfg["feature_dim"]
        self.feature_rate = self.task_cfg["feature_rate"]

        self.ast_kernel_size_chan = self.cfg["ast_kernel_size_chan"]
        self.ast_kernel_size_time = self.cfg["ast_kernel_size_time"]
        self.ast_kernel_stride_chan = self.cfg["ast_kernel_stride_chan"]
        self.ast_kernel_stride_time = self.cfg["ast_kernel_stride_time"]
        self.num_vertical_patch = (
            self.feature_dim - self.ast_kernel_size_chan
        ) // self.ast_kernel_stride_chan + 1

        self.is_decoder_finetune = False  # TODO bad way of passing in info

    def get_downsample_rates(self, key: str = None) -> int:
        return int(self.ast_kernel_stride_time * FBANK_SAMPLE_STRIDE)

    def wav_to_spectrogram(self, wav):
        return torchaudio.compliance.kaldi.fbank(  # Frame shift and length are standard at 10, 25
            waveform=wav,
            sample_frequency=self.sample_rate,
            use_energy=False,
            num_mel_bins=self.feature_dim,
        )

    def forward(self, wavs):
        device = wavs[0].device

        wavs_len = [len(wav) for wav in wavs]
        features = [self.wav_to_spectrogram(wav.unsqueeze(0)) for wav in wavs]
        feature_lengths = torch.LongTensor([len(feature) for feature in features]).to(
            device
        )
        feature_padding_mask = ~torch.lt(
            torch.arange(max(feature_lengths)).unsqueeze(0).to(device),
            feature_lengths.unsqueeze(1),
        )
        padded_features = pad_sequence(features, batch_first=True)

        results = self.model(
            padded_features,
            padding_mask=feature_padding_mask,
            mask=False,
            features_only=True,
            is_decoder_finetune=self.is_decoder_finetune,
        )
        hidden_states = results["hidden_states"]

        finalized_hs = []
        for h in hidden_states:
            # h: (batch_size, num_horizon_patch * num_vertical_patch, hidden_size)
            h = h.reshape(h.size(0), -1, self.num_vertical_patch * h.size(-1))
            h = h[:, : len(range(0, max(wavs_len), self.get_downsample_rates())), :]
            finalized_hs.append(h)

        return {
            "hidden_states": finalized_hs,
        }



"""
Basic G2P

Authors:
  * Heng-Jui Chang 2022
"""

import logging
from collections import defaultdict
from typing import Dict, List, Tuple

from s3prl.util.download import _urls_to_filepaths

DEFAULT_LEXICON_URL = [
    "https://huggingface.co/datasets/s3prl/g2p/raw/main/lexicon/librispeech-lexicon-200k-g2p.txt",
    "https://huggingface.co/datasets/s3prl/g2p/raw/main/lexicon/librispeech-lexicon-allothers-g2p.txt",
]

__all__ = ["G2P"]


def parse_lexicon(line: str) -> Tuple[str, List[str]]:
    line.replace("\t", " ")
    word, *phonemes = line.split()
    return word, phonemes


def read_lexicon_files(file_list: List[str]) -> Dict[str, List[str]]:
    w2p_dict = defaultdict(list)
    for file in file_list:
        with open(file, "r") as fp:
            lines = [line.strip() for line in fp]
            for line in lines:
                word, phonemes = parse_lexicon(line)
                w2p_dict[word].append(phonemes)

    w2p = {}
    for word, phonemes_all in w2p_dict.items():
        if len(phonemes_all) > 1:
            logging.info(f"{len(phonemes_all)} phoneme sequences found for {word}.")
            for i, phonemes in enumerate(phonemes_all):
                logging.info(f"{i}. {phonemes}")
        w2p[word] = phonemes_all[0]
    logging.info("Taking the first phoneme sequences for a deterministic behavior.")

    return w2p


class G2P:
    """Grapheme-to-phoneme

    Args:
        file_list (List[str], optional): List of lexicon files. Defaults to None.
        allow_unk (bool): If false, raise Error when a word can not be recognized by this basic G2P
    """

    def __init__(self, file_list: List[str] = None, allow_unk: bool = False):
        self.allow_unk = allow_unk

        if file_list is None:
            file_list = _urls_to_filepaths(*DEFAULT_LEXICON_URL)
        self.word2phone = read_lexicon_files(file_list)

    def encode(self, text: str) -> str:
        """Converts grapheme-based sentences to phonemes

        Args:
            text (str): Sentence

        Returns:
            str: Phonemized sentence
        """

        word_list = text.strip().upper().split(" ")
        phonemes = []
        for word in word_list:
            if not self.allow_unk:
                assert word in self.word2phone
            phonemes += self.word2phone.get(word, ["<UNK>"])

        return " ".join(phonemes)



import logging
import os
import shutil
import tempfile
import traceback
from pathlib import Path
from subprocess import check_call

import pytest
import torch
from filelock import FileLock

from s3prl.nn import Featurizer, S3PRLUpstream
from s3prl.util.download import _urls_to_filepaths
from s3prl.util.pseudo_data import get_pseudo_wavs

logger = logging.getLogger(__name__)

TEST_MORE_ITER = 2
TRAIN_MORE_ITER = 5
SAMPLE_RATE = 16000
ATOL = 0.01
MAX_LENGTH_DIFF = 3
EXTRA_SHORT_SEC = 0.05
EXTRACTED_GT_DIR = Path(__file__).parent.parent / "sample_hidden_states"
S3PRL_HF_SAMPLE_HS = "https://huggingface.co/datasets/s3prl/sample_hidden_states"

# Expect the following directory structure:
#
# -- s3prl/  (repository root)
# ---- s3prl/  (package root)
# ---- test/
# ------- test_upstream.py
# ---- sample_hidden_states/


def _prepare_sample_hidden_states():
    lock_file = Path(__file__).parent.parent / "sample_hidden_states.lock"
    with FileLock(str(lock_file)):
        # NOTE: home variable is necessary for git lfs to work
        env = dict(os.environ)
        if not "HOME" in env:
            env["HOME"] = str(Path.home())

        if not EXTRACTED_GT_DIR.is_dir():
            with tempfile.TemporaryDirectory() as tempdir:
                tempdir = Path(tempdir)
                tempdir.mkdir(exist_ok=True, parents=True)

                logger.info("Downloading extracted sample hidden states...")
                check_call("git lfs install".split(), cwd=tempdir, env=env)
                check_call(
                    f"git clone {S3PRL_HF_SAMPLE_HS}".split(),
                    cwd=tempdir,
                    env=env,
                )
                shutil.move(
                    str(tempdir / "sample_hidden_states"), str(EXTRACTED_GT_DIR.parent)
                )
        else:
            logger.info(f"{EXTRACTED_GT_DIR} exists. Perform git pull...")
            check_call(
                f"git pull {S3PRL_HF_SAMPLE_HS} main".split(),
                cwd=EXTRACTED_GT_DIR,
                env=env,
            )

    try:
        lock_file.unlink()
    except FileNotFoundError:
        pass


def _extract_feat(
    model: S3PRLUpstream,
    seed: int = 0,
    **pseudo_wavs_args,
):
    wavs, wavs_len = get_pseudo_wavs(seed=seed, padded=True, **pseudo_wavs_args)
    all_hs, all_lens = model(wavs, wavs_len)
    return all_hs


def _all_hidden_states_same(hs1, hs2):
    for h1, h2 in zip(hs1, hs2):
        if h1.size(1) != h2.size(1):
            length_diff = abs(h1.size(1) - h2.size(1))
            assert length_diff <= MAX_LENGTH_DIFF, f"{length_diff} > {MAX_LENGTH_DIFF}"
            min_seqlen = min(h1.size(1), h2.size(1))
            h1 = h1[:, :min_seqlen, :]
            h2 = h2[:, :min_seqlen, :]
            assert torch.allclose(h1, h2, atol=ATOL)


def _load_ground_truth(name: str):
    source = f"{EXTRACTED_GT_DIR}/{name}.pt"
    if source.startswith("http"):
        paths = _urls_to_filepaths(source)
        if isinstance(paths, list):
            path = paths[0]
        else:
            path = paths
    else:
        path = source
    return torch.load(path)


def _compare_with_extracted(name: str):
    model = S3PRLUpstream(name)
    model.eval()

    with torch.no_grad():
        hs = _extract_feat(model)
        hs_gt = _load_ground_truth(name)

        _all_hidden_states_same(hs, hs_gt)

        for i in range(TEST_MORE_ITER):
            more_hs = _extract_feat(model)
            for h1, h2 in zip(hs, more_hs):
                assert torch.allclose(
                    h1, h2
                ), "should have deterministic representation in eval mode"

        for i in range(TEST_MORE_ITER):
            more_hs = _extract_feat(model, seed=i + 1)
            assert len(hs) == len(
                more_hs
            ), "should have deterministic num_layer in eval mode"

    model.train()
    for i in range(TRAIN_MORE_ITER):
        more_hs = _extract_feat(model, seed=i + 1)
        assert len(hs) == len(
            more_hs
        ), "should have deterministic num_layer in train mode"


def _test_forward_backward(name: str, **pseudo_wavs_args):
    """
    Test the upstream with the name: 'name' can successfully forward and backward
    """
    with torch.autograd.set_detect_anomaly(True):
        model = S3PRLUpstream(name)
        hs = _extract_feat(model, **pseudo_wavs_args)
        h_sum = torch.zeros(1)
        for h in hs:
            h_sum: torch.Tensor = h_sum + h.sum()
        h_sum.backward()


def _filter_options(options: list):
    options = [
        name
        for name in options
        if (not name == "customized_upstream")
        and (
            not "mos" in name
        )  # mos models do not have hidden_states key. They only return a single mos score
        and (
            not "stft_mag" in name
        )  # stft_mag upstream must past the config file currently and is not so important. So, skip the test now
        and (
            not "pase" in name
        )  # pase_plus needs lots of dependencies and is difficult to be tested and is not very worthy today
        and (
            not name == "xls_r_1b"
        )  # skip due to too large model, too long download time
        and (
            not name == "xls_r_2b"
        )  # skip due to too large model, too long download time
        and (
            not name in ["ast", "ssast_patch_base", "ssast_frame_base"]
        )  # FIXME: remove timm dependency
        and (not name == "vggish")  # FIXME: remove resampy dependency
        and (not name == "byol_s_cvt")  # FIXME: remove einops dependency
        and (not "lighthubert" in name)  # FIXME: solve the random subnet issue
        and (not name == "passt_hop160base2lvl")  # too huge memory usage
        and (not name == "passt_hop160base2lvlmel")  # too huge memory usage
        and (not name == "passt_hop100base2lvl")  # too huge memory usage
        and (not name == "passt_hop100base2lvlmel")  # too huge memory usage
    ]
    options = [option for option in options if "passt" in option]
    return options


"""
Test cases ensure that all upstreams are working and are same with pre-extracted features
"""


def _test_specific_upstream(name: str):
    _compare_with_extracted(name)
    _test_forward_backward(
        name, min_secs=EXTRA_SHORT_SEC, max_secs=EXTRA_SHORT_SEC, n=1
    )
    _test_forward_backward(
        name, min_secs=EXTRA_SHORT_SEC, max_secs=EXTRA_SHORT_SEC, n=2
    )
    _test_forward_backward(name, min_secs=EXTRA_SHORT_SEC, max_secs=1, n=3)


@pytest.mark.upstream
@pytest.mark.parametrize(
    "name",
    [
        "apc",
        "audio_albert",
        "fbank",
        "mel",
        "modified_cpc",
        "data2vec",
        "decoar_layers",
        "decoar2",
        "distilhubert",
        # "espnet_hubert_base_iter1",  # espnet will be tested separately due to complex dependency
        "hubert",
        "lighthubert_base",
        "mockingjay",
        "npc",
        "discretebert",
        "tera",
        "unispeech_sat_base",
        "vq_apc",
        "vq_wav2vec",
        "wav2vec",
        "wav2vec2",
        "wavlm",
    ],
)
def test_common_upstream(name):
    if "espnet" in name:
        try:
            import espnet
        except:
            logger.info("Skip ESPNet upstream test cases if espnet is not installed")
            return

    _prepare_sample_hidden_states()
    _test_specific_upstream(name)


@pytest.mark.upstream
@pytest.mark.parametrize(
    "name",
    [
        "apc",
        "audio_albert",
        "decoar2",
        "distilhubert",
        # "espnet_hubert_base_iter1",  # espnet will be tested separately due to complex dependency
        "hubert",
        "mockingjay",
    ],
)
def test_few_upstream(name):
    if "espnet" in name:
        try:
            import espnet
        except:
            logger.info("Skip ESPNet upstream test cases if espnet is not installed")
            return

    _prepare_sample_hidden_states()
    _test_specific_upstream(name)


@pytest.mark.upstream
def test_specific_upstream(upstream_names: str):
    _prepare_sample_hidden_states()
    if upstream_names is not None:
        options = upstream_names.split(",")

        tracebacks = []
        for name in options:
            logger.info(f"Testing upstream: '{name}'")
            try:
                _test_specific_upstream(name)
            except Exception as e:
                logger.error(f"{name}\n{traceback.format_exc()}")
                tb = traceback.format_exc()
                tracebacks.append((name, tb))

        if len(tracebacks) > 0:
            for name, tb in tracebacks:
                logger.error(f"Error in {name}:\n{tb}")
            logger.error(f"All failed models:\n{[name for name, _ in tracebacks]}")
            assert False


@pytest.mark.upstream
@pytest.mark.slow
def test_upstream_with_extracted(upstream_names: str):
    _prepare_sample_hidden_states()

    if upstream_names is not None:
        options = upstream_names.split(",")
    else:
        options = S3PRLUpstream.available_names(only_registered_ckpt=True)
        options = _filter_options(options)
        options = sorted(options)

    tracebacks = []
    for name in options:
        logger.info(f"Testing upstream: '{name}'")
        try:
            _compare_with_extracted(name)

        except Exception as e:
            logger.error(f"{name}\n{traceback.format_exc()}")
            tb = traceback.format_exc()
            tracebacks.append((name, tb))

    if len(tracebacks) > 0:
        for name, tb in tracebacks:
            logger.error(f"Error in {name}:\n{tb}")
        logger.error(f"All failed models:\n{[name for name, _ in tracebacks]}")
        assert False


@pytest.mark.upstream
@pytest.mark.slow
def test_upstream_forward_backward(upstream_names: str):
    if upstream_names is not None:
        options = upstream_names.split(",")
    else:
        options = S3PRLUpstream.available_names(only_registered_ckpt=True)
        options = _filter_options(options)
        options = sorted(options)
        options = reversed(options)

    tracebacks = []
    for name in options:
        logger.info(f"Testing upstream: '{name}'")
        try:
            _test_forward_backward(name)

        except Exception as e:
            logger.error(f"{name}\n{traceback.format_exc()}")
            tb = traceback.format_exc()
            tracebacks.append((name, tb))

    if len(tracebacks) > 0:
        for name, tb in tracebacks:
            logger.error(f"Error in {name}:\n{tb}")
        logger.error(f"All failed models:\n{[name for name, _ in tracebacks]}")
        assert False


@pytest.mark.upstream
@pytest.mark.parametrize("layer_selections", [None, [0, 4, 9]])
@pytest.mark.parametrize("normalize", [False, True])
def test_featurizer(layer_selections, normalize):
    model = S3PRLUpstream("hubert")
    featurizer = Featurizer(
        model, layer_selections=layer_selections, normalize=normalize
    )

    wavs, wavs_len = get_pseudo_wavs(padded=True)
    all_hs, all_lens = model(wavs, wavs_len)
    hs, hs_len = featurizer(all_hs, all_lens)

    assert isinstance(hs, torch.FloatTensor)
    assert isinstance(hs_len, torch.LongTensor)


@pytest.mark.upstream
def test_upstream_properties():
    model = S3PRLUpstream("hubert")
    featurizer = Featurizer(model)
    assert isinstance(model.hidden_sizes, (tuple, list)) and isinstance(
        model.hidden_sizes[0], int
    )
    assert isinstance(model.downsample_rates, (tuple, list)) and isinstance(
        model.downsample_rates[0], int
    )
    assert isinstance(featurizer.output_size, int)
    assert isinstance(featurizer.downsample_rate, int)



from typing import Dict, List, Union

import torch.nn as nn
from torch import Tensor
from torch.nn.utils.rnn import pad_sequence

from ..interfaces import SAMPLE_RATE
from .serab_byols import serab


class UpstreamExpert(nn.Module):
    def __init__(
        self,
        ckpt: str = None,
        model_name: str = None,
        window_secs: float = 1,
        hop_secs: float = 0.05,
        model_config: str = None,
    ):
        super().__init__()
        self.model = serab.load_model(ckpt, model_name)
        self.frame_duration = window_secs * 1000
        self.hop_size = hop_secs * 1000
        self.model_config = model_config

    def get_downsample_rates(self, key: str = None) -> int:
        return int(self.hop_size / 1000 * SAMPLE_RATE)

    def forward(self, wavs: List[Tensor]) -> Dict[str, Union[Tensor, List[Tensor]]]:
        padded_wavs = pad_sequence(wavs, batch_first=True)
        embeddings, timestamps = serab.get_timestamp_embeddings(
            padded_wavs, self.model, self.frame_duration, self.hop_size
        )
        return {
            "hidden_states": [embeddings],
        }



"""
HEAR Competition submission script following the
https://neuralaudio.ai/hear2021-holistic-evaluation-of-audio-representations.html#common-api
guidelines
"""

from pathlib import Path
from typing import List, Tuple

import torch
from torch import Tensor
from torchaudio.transforms import MelSpectrogram

from ..byol_a.augmentations import PrecomputedNorm
from ..byol_a.common import load_yaml_config
from ..byol_a.models.audio_ntt import AudioNTT2020
from ..byol_a.models.clstm import CLSTM
from ..byol_a.models.cvt import CvT
from ..byol_a.models.resnetish import resnetish34
from .utils import *

# Default frame duration in milliseconds
TIMESTAMP_FRAME_DUR = 1000
# Default hop_size in milliseconds
TIMESTAMP_HOP_SIZE = 50

# Number of frames to batch process for timestamp embeddings
BATCH_SIZE = 512


def get_model(model_name: str = "", cfg={}) -> torch.nn.Module:
    """Define the model object.

    Parameters
    ----------
    model_name: str, the name for pretrained model
    cfg: dict, the cfg parameters

    Returns
    -------
    torch.nn.Module object or a tensorflow "trackable" object
    """
    if model_name == "default":
        model = AudioNTT2020(n_mels=cfg.n_mels, d=cfg.feature_d)

    elif model_name == "resnetish34":
        model = resnetish34()

    elif model_name == "clstm":
        model = CLSTM()

    elif model_name == "cvt":
        s1_depth, s2_depth, s3_depth = cfg.depths
        s1_emb_dim, s2_emb_dim, s3_emb_dim = cfg.embed_dims
        s1_mlp_mult, s2_mlp_mult, s3_mlp_mult = cfg.mlp_mults

        model = CvT(
            s1_emb_dim=s1_emb_dim,
            s1_depth=s1_depth,
            s1_mlp_mult=s1_mlp_mult,
            s2_emb_dim=s2_emb_dim,
            s2_depth=s2_depth,
            s2_mlp_mult=s2_mlp_mult,
            s3_emb_dim=s3_emb_dim,
            s3_depth=s3_depth,
            s3_mlp_mult=s3_mlp_mult,
            pool=cfg.cvt_pool,
        )
    else:
        raise ValueError("Model not found.")
    return model


def load_model(
    model_file_path: str = "",
    model_name: str = "default",
    cfg_path: str = None,
) -> torch.nn.Module:
    """Load pre-trained DL models.

    Parameters
    ----------
    model_name: str, the name for pretrained model
    model_file_path: str, the path for pretrained model
    cfg_path: str, the path for yaml file including parameters value

    Returns
    -------
    torch.nn.Module object or a tensorflow "trackable" object
        Model loaded with pre-training weights
    """
    cfg_path = cfg_path or Path(__file__).parent / "config.yaml"
    # assert model_name in model_file_path.split('_')[0], "The checkpoint doesn't match with the selected model name"

    # Load config file
    cfg = load_yaml_config(cfg_path)

    # Load pretrained weights.
    model = get_model(model_name, cfg)

    state_dict = torch.load(model_file_path)
    model.load_state_dict(state_dict)
    return model


def get_timestamp_embeddings(
    audio_list: List,
    model: torch.nn.Module,
    frame_duration: float = TIMESTAMP_FRAME_DUR,
    hop_size: float = TIMESTAMP_HOP_SIZE,
    cfg_path: str = None,
) -> Tuple[Tensor, Tensor]:
    """
    This function returns embeddings at regular intervals centered at timestamps. Both
    the embeddings and corresponding timestamps (in milliseconds) are returned.
    Args:
        audio_list: List of torch tensor audios.
        model: Loaded model.
        frame_duration: Frame (segement) duration in milliseconds
        hop_size: Hop size in milliseconds.
            NOTE: Not required by the HEAR API. We add this optional parameter
            to improve the efficiency of scene embedding.
        cfg_path: str, the path for yaml file including parameters value
    Returns:
        - Tensor: embeddings, A float32 Tensor with shape (n_sounds, n_timestamps,
            model.timestamp_embedding_size).
        - Tensor: timestamps, Centered timestamps in milliseconds corresponding
            to each embedding in the output. Shape: (n_sounds, n_timestamps).
    """
    cfg_path = cfg_path or Path(__file__).parent / "config.yaml"

    # Load config file
    cfg = load_yaml_config(cfg_path)
    to_melspec = MelSpectrogram(
        sample_rate=cfg.sample_rate,
        n_fft=cfg.n_fft,
        win_length=cfg.win_length,
        hop_length=cfg.hop_length,
        n_mels=cfg.n_mels,
        f_min=cfg.f_min,
        f_max=cfg.f_max,
    ).to(audio_list[0].device)

    # Send the model to the same device that the audio tensor is on.
    model = model.to(audio_list[0].device)

    # Split the input audio signals into frames and then flatten to create a tensor
    # of audio frames that can be batch processed.
    frames, timestamps = frame_audio(
        audio_list,
        frame_size=(frame_duration / 1000) * cfg.sample_rate,
        hop_size=hop_size,
        sample_rate=cfg.sample_rate,
    )
    audio_batches, num_frames, _ = frames.shape
    frames = frames.flatten(end_dim=1)
    # (batch_size * num_window, window_size)

    # Convert audio frames to Log Mel-spectrograms
    melspec_frames = (to_melspec(frames) + torch.finfo(torch.float).eps).log()
    # (batch_size * num_window, fbank_size, num_frames)

    normalizer = PrecomputedNorm(compute_timestamp_stats(melspec_frames))
    melspec_frames = normalizer(melspec_frames).unsqueeze(1)
    # (batch_size * num_window, 1, fbank_size, num_frames)

    embeddings = model(melspec_frames)
    embeddings = embeddings.reshape(audio_batches, num_frames, -1)

    return embeddings, timestamps


def get_scene_embeddings(
    audio_list: List,
    model: torch.nn.Module,
    cfg_path: str = None,
) -> Tensor:
    """
    This function returns a single embedding for each audio clip. In this baseline
    implementation we simply summarize the temporal embeddings from
    get_timestamp_embeddings() using torch.mean().
    Args:
        audio_list: list of torch tensor audios (audios should be resampled to 16kHz).
        model: Loaded model.
        cfg_path:
    Returns:
        - embeddings, A float32 Tensor with shape
            (n_sounds, model.scene_embedding_size).
    """
    cfg_path = cfg_path or Path(__file__).parent / "config.yaml"

    device = audio_list[0].device
    # Load config file
    cfg = load_yaml_config(cfg_path)
    to_melspec = MelSpectrogram(
        sample_rate=cfg.sample_rate,
        n_fft=cfg.n_fft,
        win_length=cfg.win_length,
        hop_length=cfg.hop_length,
        n_mels=cfg.n_mels,
        f_min=cfg.f_min,
        f_max=cfg.f_max,
    ).to(device)
    stats = compute_scene_stats(audio_list, to_melspec)
    normalizer = PrecomputedNorm(stats)
    model = model.to(device)
    embeddings = generate_byols_embeddings(
        model, audio_list, to_melspec, normalizer, device
    )
    return embeddings



"""
Parse the IEMOCAP corpus

Authors:
  * Leo 2022
  * Cheng Liang 2022
"""

import logging
import re
from copy import deepcopy
from pathlib import Path

from librosa.util import find_files

from .base import Corpus

IEMOCAP_SESSION_NUM = 5
LABEL_DIR_PATH = "dialog/EmoEvaluation"
WAV_DIR_PATH = "sentences/wav"

logger = logging.getLogger(__name__)

__all__ = [
    "IEMOCAP",
]


class IEMOCAP(Corpus):
    """
    Parse the IEMOCAP dataset

    Args:
        dataset_root: (str) The dataset root of IEMOCAP
    """

    def __init__(self, dataset_root: str, n_jobs: int = 4) -> None:
        self.dataset_root = Path(dataset_root)
        self.sessions = [
            (self._preprocess_single_session(self.dataset_root, session_id))
            for session_id in range(1, IEMOCAP_SESSION_NUM + 1)
        ]

        self._all_data = dict()
        for session in self.sessions:
            self._all_data.update(session["improvised"])
            self._all_data.update(session["scripted"])

    @staticmethod
    def _preprocess_single_session(dataset_root: Path, session_id: int):
        data = dict(
            improvised={},
            scripted={},
        )

        session_dir = dataset_root / f"Session{session_id}"
        label_dir = session_dir / LABEL_DIR_PATH
        wav_root_dir = session_dir / WAV_DIR_PATH
        wav_paths = find_files(wav_root_dir)
        for wav_path in wav_paths:
            wav_path = Path(wav_path)
            spk_and_act_and_scene = wav_path.parts[-2]
            label_file = label_dir / f"{spk_and_act_and_scene}.txt"
            with label_file.open() as file:
                content = file.read()
            result = re.search(rf"{str(wav_path.stem)}\t(.+)\t", content)
            speaker = spk_and_act_and_scene.split("_")[0]
            act = "improvised" if "impro" in spk_and_act_and_scene else "scripted"
            emotion = result.groups()[0]
            unique_id = wav_path.stem

            data[act][unique_id] = dict(
                wav_path=str(wav_path),
                speaker=speaker,
                act=act,
                emotion=emotion,
                session_id=session_id,
            )

        return data

    @property
    def all_data(self):
        """
        Return:
            dict

            all the data points of IEMOCAP in the format of

            .. code-block:: yaml

                data_id1:
                    wav_path (str): The waveform path
                    speaker (str): The speaker name
                    act (str): improvised / scripted
                    emotion (str): The emotion label
                    session_id (int): The session

                data_id2:
                    ...
        """
        return deepcopy(self._all_data.copy())

    def get_whole_session(self, session_id: int):
        """
        Args:
            session_id (int): The session index selected from 1, 2, 3, 4, 5

        Return:
            dict

            data points in a single session (containing improvised and scripted recordings) in the
            same format as :obj:`all_data`
        """
        output = dict()
        output.update(self.get_session_with_act(session_id, "improvised"))
        output.update(self.get_session_with_act(session_id, "scripted"))
        return deepcopy(output)

    def get_session_with_act(self, session_id: int, act: str):
        """
        Args:
            session_id (int): The session index selected from 1, 2, 3, 4, 5
            act (str): 'improvised' or 'scripted'

        Return:
            :obj:`s3prl.base.container.Container`

            data points in a single session with a specific act (either improvised or scripted) in the
            same format as :obj:`all_data`
        """
        assert act in ["improvised", "scripted"]
        return deepcopy(self.sessions[session_id - 1][act])

    @classmethod
    def download_dataset(cls, tgt_dir: str) -> None:
        import os
        import tarfile

        import requests

        assert os.path.exists(
            os.path.abspath(tgt_dir)
        ), "Target directory does not exist"

        def unzip_targz_then_delete(filepath: str):
            with tarfile.open(os.path.abspath(filepath)) as tar:
                tar.extractall(path=os.path.abspath(tgt_dir))
            os.remove(os.path.abspath(filepath))

        def download_from_url(url: str):
            filename = url.split("/")[-1].replace(" ", "_")
            filepath = os.path.join(tgt_dir, filename)

            r = requests.get(url, stream=True)
            if r.ok:
                logger.info(f"Saving {filename} to", os.path.abspath(filepath))
                with open(filepath, "wb") as f:
                    for chunk in r.iter_content(chunk_size=1024 * 1024 * 10):
                        if chunk:
                            f.write(chunk)
                            f.flush()
                            os.fsync(f.fileno())
                logger.info(f"{filename} successfully downloaded")
                unzip_targz_then_delete(filepath)
            else:
                logger.info(f"Download failed: status code {r.status_code}\n{r.text}")

        if not os.path.exists(os.path.join(os.path.abspath(tgt_dir), "IEMOCAP/")):
            download_from_url("http://140.112.21.28:9000/IEMOCAP.tar.gz")
        logger.info(
            f"IEMOCAP dataset downloaded. Located at {os.path.abspath(tgt_dir)}/IEMOCAP/"
        )



"""
Utility functions for hear-kit
"""
from typing import Tuple

import numpy as np
import torch
import torch.nn.functional as F
from torch import Tensor
from tqdm import tqdm


def compute_scene_stats(audios, to_melspec):
    mean = 0.0
    std = 0.0
    for audio in audios:
        # Compute log-mel-spectrogram
        lms = (to_melspec(audio) + torch.finfo(torch.float).eps).log()

        # Compute mean, std
        mean += lms.mean()
        std += lms.std()

    mean /= len(audios)
    std /= len(audios)
    stats = [mean.item(), std.item()]
    return stats


def compute_timestamp_stats(melspec):
    """Compute statistics of the mel-spectrograms.

    Parameters
    ----------
    melspec : Tensor of shape (n_sounds*n_frames, n_mels, time)

    Returns
    -------
    list containing the mean and the standard deviation of the mel-spectrograms"""
    # Compute mean, std
    mean = melspec.mean()
    std = melspec.std()
    mean /= len(melspec)
    std /= len(melspec)

    stats = [mean.item(), std.item()]
    return stats


def frame_audio(
    audio: Tensor, frame_size: int, hop_size: float, sample_rate: int
) -> Tuple[Tensor, Tensor]:
    """
    Adapted from https://github.com/neuralaudio/hear-baseline/hearbaseline/

    Slices input audio into frames that are centered and occur every
    sample_rate * hop_size samples. We round to the nearest sample.
    Args:
        audio: input audio, expects a 2d Tensor of shape:
            (n_sounds, num_samples)
        frame_size: the number of samples each resulting frame should be
        hop_size: hop size between frames, in milliseconds
        sample_rate: sampling rate of the input audio
    Returns:
        - A Tensor of shape (n_sounds, num_frames, frame_size)
        - A Tensor of timestamps corresponding to the frame centers with shape:
            (n_sounds, num_frames).
    """

    # Zero pad the beginning and the end of the incoming audio with half a frame number
    # of samples. This centers the audio in the middle of each frame with respect to
    # the timestamps.
    frame_size = int(frame_size)
    audio = F.pad(audio, (frame_size // 2, frame_size - frame_size // 2))
    num_padded_samples = audio.shape[1]

    frame_step = int(hop_size / 1000.0 * sample_rate)
    frame_number = 0
    frames = []
    timestamps = []
    frame_start = 0
    frame_end = frame_size
    while True:
        frames.append(audio[:, frame_start:frame_end])
        timestamps.append(frame_number * frame_step / sample_rate * 1000.0)

        # Increment the frame_number and break the loop if the next frame end
        # will extend past the end of the padded audio samples
        frame_number += 1
        frame_start = int(round(frame_number * frame_step))
        frame_end = frame_start + frame_size

        if not frame_end <= num_padded_samples:
            break

    # Expand out the timestamps to have shape (n_sounds, num_frames)
    timestamps_tensor = torch.tensor(timestamps, dtype=torch.float32)
    timestamps_tensor = timestamps_tensor.expand(audio.shape[0], -1)

    return torch.stack(frames, dim=1), timestamps_tensor


def generate_byols_embeddings(model, audios, to_melspec, normalizer, device):
    """
    Generate audio embeddings from a pretrained feature extractor.

    Converts audios to float, resamples them to the desired learning_rate,
    and produces the embeddings from a pre-trained model.

    Adapted from https://github.com/google-research/google-research/tree/master/non_semantic_speech_benchmark

    Parameters
    ----------
    model : torch.nn.Module object or a tensorflow "trackable" object
        Model loaded with pre-training weights
    audios : list
        List of audios, loaded as a numpy arrays
    to_melspec : torchaudio.transforms.MelSpectrogram object
        Mel-spectrogram transform to create a spectrogram from an audio signal
    normalizer : nn.Module
        Pre-normalization transform
    device : torch.device object
        Used device (CPU or GPU)

    Returns
    ----------
    embeddings: Tensor
        2D Array of embeddings for each audio of size (N, M). N = number of samples, M = embedding dimension
    """
    embeddings = []
    model.eval()
    for param in model.parameters():
        param.requires_grad = False
    with torch.no_grad():
        for audio in tqdm(audios, desc=f"Generating Embeddings...", total=len(audios)):
            lms = normalizer(
                (
                    to_melspec(audio.to(device).unsqueeze(0))
                    + torch.finfo(torch.float).eps
                ).log()
            ).unsqueeze(0)
            embedding = model(lms)
            embeddings.append(embedding)
    embeddings = torch.cat(embeddings, dim=0)
    return embeddings



# --------------------------------------------------------
# LightHuBERT: Lightweight and Configurable Speech Representation Learning with Once-for-All Hidden-Unit BERT (https://arxiv.org/pdf/2203.15610.pdf)
# Github source: https://github.com/mechanicalsea/lighthubert
# Licensed under The MIT License [see LICENSE for details]
# --------------------------------------------------------

import numbers

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor


class SLayerNorm(nn.LayerNorm):
    """LayerNorm: variable 1-D size
    __base__: torch.nn.LayerNorm
    """

    def __init__(
        self, normalized_shape: int, eps: float = 1e-5, elementwise_affine: bool = True
    ) -> None:
        super(SLayerNorm, self).__init__(normalized_shape, eps, elementwise_affine)
        self.staticize()

    def staticize(self):
        self.sample_normalized_shape = self.normalized_shape[0]
        self.samples = {
            "weight": self.weight,
            "bias": self.bias,
        }

    def set_sample_config(self, sample_normalized_shape: int):
        self.sample_normalized_shape = sample_normalized_shape
        self._sample_parameters()

    def _sample_parameters(self):
        if self.elementwise_affine:
            self.samples["weight"] = self.weight[: self.sample_normalized_shape]
            self.samples["bias"] = self.bias[: self.sample_normalized_shape]
        else:
            self.samples["weight"] = None
            self.samples["bias"] = None
        return self.samples

    def calc_sampled_param_num(self):
        return self.samples["weight"].numel() + self.samples["bias"].numel()

    def get_complexity(self, sequence_length):
        return sequence_length * self.sample_normalized_shape

    @property
    def weights(self):
        return self.samples["weight"] if self.elementwise_affine else None

    @property
    def biases(self):
        return self.samples["bias"] if self.elementwise_affine else None

    @property
    def normalized_shapes(self):
        if isinstance(self.sample_normalized_shape, numbers.Integral):
            # mypy error: incompatible types in assignment
            sample_normalized_shape = (self.sample_normalized_shape,)  # type: ignore[assignment]
        else:
            sample_normalized_shape = self.sample_normalized_shape
        return tuple(sample_normalized_shape)  # type: ignore[arg-type]

    def forward(self, input: Tensor) -> Tensor:
        self._sample_parameters()
        return F.layer_norm(
            input, self.normalized_shapes, self.weights, self.biases, self.eps
        )

    def extra_repr(self) -> str:
        return (
            f"{self.normalized_shape}, eps={self.eps}, "
            f"elementwise_affine={self.elementwise_affine}"
        )

    def clone_model(self, normalized_shape: int):
        self.set_sample_config(normalized_shape)

        m = nn.LayerNorm(normalized_shape, self.eps, self.elementwise_affine)
        if m.elementwise_affine:
            m = m.to(self.weight.device)
            m = m.to(self.weight.dtype)
            m.weight.data.copy_(self.weights)
            m.bias.data.copy_(self.biases)
        return m.eval()

    @classmethod
    def build_from(cls, m: nn.LayerNorm):
        normalized_shape = m.normalized_shape
        eps = m.eps
        elementwise_affine = m.elementwise_affine
        _m = cls(normalized_shape, eps, elementwise_affine)
        if _m.elementwise_affine:
            _m = _m.to(m.weight.device)
            _m = _m.to(m.weight.dtype)
            _m.weight.data.copy_(m.weight)
            _m.bias.data.copy_(m.bias)
        return _m


if __name__ == "__main__":
    m = SLayerNorm(3)
    print(m)
    print(m.weight.data, m.bias.data)
    m.set_sample_config(2)
    z = m.clone_model(2)
    x = SLayerNorm.build_from(z)
    print(m)
    print(m.weight.data, m.bias.data)
    print(z)
    print(z.weight.data, z.bias.data)
    print(x)
    print(x.weight.data, x.bias.data)
    inp = torch.rand((1, 2))
    print(torch.allclose(m(inp), z(inp)))
    print(torch.allclose(m(inp), x(inp)))



# --------------------------------------------------------
# LightHuBERT: Lightweight and Configurable Speech Representation Learning with Once-for-All Hidden-Unit BERT (https://arxiv.org/pdf/2203.15610.pdf)
# Github source: https://github.com/mechanicalsea/lighthubert
# Licensed under The MIT License [see LICENSE for details]
# --------------------------------------------------------

"""
Sliding window attention.
- `sliding_attn` implements it via `torch.as_strided` with reducing memory and speedup computation.
- `sliding_attn_check_mask` implements it via `attn_mask` as global attention.
- `global_attention_forward` is global attention with full-utterance sliding window size.
"""

import torch
import torch.nn.functional as F


def pad_as_attn_swz(query, key, value, swz, attn_mask=None, key_padding_mask=None):
    """
    Force the query, key, and value to be the length of the multiple of sliding window size.
        We add zero at the beginning of them but does not mask them to avoid grad nan.

    Args:
        query, key, value:
            3-d tensor (batch_size, seq_len, hidden_dim)
        swz:
            int, sliding attention window size
        attn_mask:
            2-d bool tensor (tgt_len, src_len)
            3-d bool tensor (bsz, tgt_len, src_len)
        key_padding_mask:
            2-d bool tensor (batch_size, max_seq_len)

    Returns:
        query, key, value, attn_mask, key_padding_mask, tgt_len, src_len
    """
    if (
        isinstance(swz, int)
        and query.size(1) % swz == 0
        and key.size(1) % swz == 0
        and value.size(1) % swz == 0
    ):
        return (
            query,
            key,
            value,
            attn_mask,
            key_padding_mask,
            query.size(1),
            key.size(1),
        )

    tgt_len, src_len = query.size(1), key.size(1)
    pad_tgt = (tgt_len // swz + 1) * swz - tgt_len if tgt_len % swz > 0 else 0
    pad_src = (src_len // swz + 1) * swz - src_len if src_len % swz > 0 else 0
    if attn_mask is not None:
        # assert attn_mask.dtype == torch.bool
        attn_mask = F.pad(attn_mask, (pad_src, 0, pad_tgt, 0), value=False)
    if key_padding_mask is not None:
        assert key_padding_mask.ndim == 2
        assert key_padding_mask.dtype == torch.bool
        key_padding_mask = F.pad(key_padding_mask, (pad_src, 0), value=False)
    query = F.pad(query, (0, 0, pad_tgt, 0))
    key = F.pad(key, (0, 0, pad_src, 0))
    value = F.pad(value, (0, 0, pad_src, 0))
    return (query, key, value, attn_mask, key_padding_mask, tgt_len, src_len)


@torch.no_grad()
def merge_padding_attm_mask(attn_mask, key_padding_mask, num_heads, tgt_len=None):
    r"""
    Merge `key_padding_mask` into `attn_mask`.

    Args:
        attn_mask:
            2-d bool tensor (tgt_len, src_len) or
            3-d bool tensor (batch_size * num_heads, tgt_len, src_len)
            attn_mask[batch_i, t_j, t_k] = True or False
        key_padding_mask:
            1-d tensor long (batch_size) or 2-d tensor bool (batch_size, src_len)
        num_heads:
            int, the number of heads

    Return:
        attn_mask:
            2-d bool tensor (tgt_len, src_len) or
            3-d bool tensor (batch_size * num_heads, 1, src_len) or
            3-d bool tensor (batch_size * num_heads, tgt_len, src_len)
    """
    if key_padding_mask is None:
        return attn_mask
    else:
        assert num_heads is not None
    if key_padding_mask.ndim == 1:
        key_padding_mask = mask_padding(key_padding_mask)
    bsz, src_len = key_padding_mask.size()
    key_padding_mask = (
        key_padding_mask.view(bsz, 1, 1, src_len)
        .expand(bsz, num_heads, 1, src_len)
        .reshape(bsz * num_heads, 1, src_len)
    )
    if attn_mask is None:
        if key_padding_mask.size(1) == 1:
            assert tgt_len is not None
            attn_mask = key_padding_mask.expand(bsz * num_heads, tgt_len, src_len)
        return attn_mask
    assert attn_mask.dtype == torch.bool, f"{attn_mask.dtype}"
    tgt_len = attn_mask.size(-2) if tgt_len is None else tgt_len
    if attn_mask.ndim == 2:
        attn_mask = attn_mask.unsqueeze(0).expand(bsz * num_heads, tgt_len, src_len)
    assert src_len == attn_mask.size(-1)
    key_padding_mask = key_padding_mask.expand(-1, tgt_len, src_len)
    attn_mask = attn_mask.logical_or(key_padding_mask)
    return attn_mask


def mask_padding(seq_len: torch.LongTensor, max_len=None):
    r"""
    Args:
        seq_len: 1-d long tensor (batch_size)

    Returns:
        key_padding_mask: 2-d bool tensor
    """
    max_len = max(seq_len) if max_len is None else max_len
    key_padding_mask = torch.arange(max_len, device=seq_len.device).unsqueeze(0)
    key_padding_mask = key_padding_mask.expand(seq_len.size(0), -1)
    return key_padding_mask >= seq_len.unsqueeze(1)


def size_check(q, k, v):
    bsz, tgt_len, src_len, head_dim = q.size(0), q.size(1), k.size(1), q.size(2)
    assert q.ndim == k.ndim == v.ndim == 3
    assert (
        q.size(0) == k.size(0) == v.size(0)
    ), f"{q.size(0)} ?= {k.size(0)} ?= {v.size(0)}"
    assert tgt_len <= src_len, f"tgt_len {tgt_len} > src_len {src_len}"
    assert k.size(1) == v.size(1), f"seq_len: {k.size(1)} != {v.size(1)}"
    assert (
        q.size(2) == k.size(2) == v.size(2)
    ), f"{q.size(1)} != {k.size(1)} != {v.size(1)}"
    return bsz, tgt_len, src_len, head_dim


def sliding_attn(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    swz: int,
    attn_mask=None,
    dropout_p=0,
    training=False,
):
    r"""
    Sliding window attention.

    Args:
        q, k, v:
            3-dim tensor, e.g., (batch_size * head_num, seq_len, head_dim)
            q is a scaled q by sqrt(head_dim).
        swz:
            int, swz % 2 == 0, sliding window size
        attn_mask:
            2-d bool tensor (tgt_len, src_len) or
            3-d bool tensor (bsz, tgt_len, src_len)

    Notes:
        `torch.as_strided` does not copy data, like `torch.Tensor.view`.
        Pad K matrix with value of -inf, but pad V matrix with value of 0.
    """
    bsz, tgt_len, src_len, head_dim = size_check(q, k, v)
    assert tgt_len == src_len and tgt_len % swz == 0, f"{tgt_len}, {src_len}, {swz}"

    # q * k
    csz = tgt_len // (swz // 2) - 1
    q_chunk = q.view(bsz, tgt_len // swz, swz, head_dim).as_strided(
        size=(bsz, csz, swz, head_dim),
        stride=(tgt_len * head_dim, swz // 2 * head_dim, head_dim, 1),
    )
    k_chunk = k.view(bsz, src_len // swz, swz, head_dim).as_strided(
        size=(bsz, csz, swz, head_dim),
        stride=(src_len * head_dim, swz // 2 * head_dim, head_dim, 1),
    )
    weights_chunk = torch.einsum(
        "bcxd,bcyd->bcxy", q_chunk, k_chunk
    )  # (bsz, csz, swz, swz)
    chunk1 = weights_chunk.as_strided(
        size=(bsz, csz, swz // 2, swz // 2 + 1),
        stride=(csz * swz * swz, swz * swz, swz + 1, 1),
    )
    chunk2 = F.pad(
        weights_chunk[:, -1, swz // 2 :, swz // 2 :],
        pad=(0, swz // 2),
        value=float("-inf"),
    ).as_strided(
        size=(bsz, swz // 2, swz // 2 + 1),
        stride=((swz // 2) * swz, swz + 1, 1),
    )
    chunk3 = F.pad(
        weights_chunk[:, 0, : swz // 2, : swz // 2],
        pad=(swz // 2, 0),
        value=float("-inf"),
    ).as_strided(
        size=(bsz, swz // 2, swz // 2),
        stride=((swz // 2) * swz, swz + 1, 1),
    )
    chunk4 = (
        weights_chunk[:, :, swz // 2 :, :]
        .contiguous()
        .as_strided(
            size=(bsz, csz, swz // 2, swz // 2),
            stride=(csz * (swz // 2) * swz, (swz // 2) * swz, swz + 1, 1),
        )
    )
    attn_weights = torch.cat(
        [
            torch.cat([chunk3.unsqueeze(1), chunk4], dim=1),
            torch.cat([chunk1, chunk2.unsqueeze(1)], dim=1),
        ],
        dim=3,
    )
    attn_weights = attn_weights.view(bsz, tgt_len, swz + 1)

    if attn_mask is not None:
        if attn_mask.ndim == 2:
            attn_mask = attn_mask.unsqueeze(0)
        attn_mask = F.pad(attn_mask, (swz // 2, swz // 2), value=True)
        attn_mask = attn_mask.as_strided(
            size=(1, tgt_len, swz + 1),
            stride=(tgt_len * (src_len + swz), src_len + swz + 1, 1),
        )
        attn_mask = attn_mask.to(attn_weights.dtype).masked_fill(
            attn_mask, float("-inf")
        )
        attn_weights += attn_mask

    attn_probs = F.softmax(attn_weights.to(torch.float32), dim=-1)
    attn_probs = attn_probs.to(attn_weights.dtype)
    attn_probs = F.dropout(attn_probs, p=dropout_p, training=training)

    # probs * v
    kpad_len1 = swz // 2
    kpad_len2 = max(swz // 2 - (src_len - tgt_len), 0)
    v_pad = F.pad(v, (0, 0, kpad_len1, kpad_len2), value=0)  # (bsz, tgt_len, )
    csz = tgt_len // (swz // 2) - 1
    assert tgt_len % (swz // 2) == 0, f"{tgt_len} % {swz // 2} != 0"
    attn_probs = attn_probs.view(bsz, csz + 1, swz // 2, swz + 1)
    attn_probs_pad = F.pad(
        attn_probs, (0, swz // 2 - 1)
    )  # (bsz, tgt_len // (swz // 2), swz // 2, swz + swz // 2)
    attn_probs_pad = attn_probs_pad.as_strided(
        size=(bsz, csz + 1, swz // 2, swz + swz // 2),
        stride=attn_probs_pad.stride()[:2] + (swz + swz // 2 - 1, 1),
    )
    v_size = (bsz, csz + 1, swz + swz // 2, head_dim)
    v_stride = (v_pad.stride(0), swz // 2 * v_pad.stride(1), head_dim, 1)
    v_chunk = v_pad.as_strided(v_size, v_stride)
    attn_chunk = torch.einsum("bcdw,bcwh->bcdh", attn_probs_pad, v_chunk)
    attn = attn_chunk.view(bsz, tgt_len, head_dim)

    return attn, attn_weights


def sliding_attn_check_mask(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    swz: int,
    attn_mask=None,
    dropout_p=0,
    training=False,
    ret_slide=True,
):
    r"""
    Check sliding attention dot product via sliding `attn_mask`.

    Args:
        q, k, v: tensor, (batch_size * head_num, seq_len, head_dim)
        swz: int, swz % 2 == 0, sliding window size
        attn_mask:
            2-d bool tensor (tgt_len, src_len) or
            3-d bool tensor (bsz, tgt_len, src_len)
        ret_slide:
            bool, if true, return sliding `attn_weights`

    Notes:
        The implementation keep compuation as full attention.
    """
    bsz, tgt_len, src_len, head_dim = size_check(q, k, v)
    with torch.no_grad():  # sliding attn mask
        swa_mask = torch.triu(q.new_ones(tgt_len, src_len), swz // 2 + 1) + torch.tril(
            q.new_ones(tgt_len, src_len), -(swz // 2 + 1)
        )
        swa_mask = swa_mask.unsqueeze(0).masked_fill(
            swa_mask.unsqueeze(0).to(torch.bool), float("-inf")
        )

    attn_weights = torch.bmm(q, k.transpose(1, 2))
    attn_weights += swa_mask

    if attn_mask is not None:
        if attn_mask.ndim == 2:
            attn_mask = attn_mask.unsqueeze(0)
        attn_mask = attn_mask.to(attn_weights.dtype).masked_fill(
            attn_mask, float("-inf")
        )
        attn_weights += attn_mask

    attn_probs = F.softmax(attn_weights.to(torch.float32), dim=-1)
    attn_probs = attn_probs.to(attn_weights.dtype)
    attn_probs = F.dropout(attn_probs, p=0, training=False)
    attn = torch.bmm(attn_probs, v)

    if not ret_slide:
        return attn, attn_weights

    kpad_len1 = swz // 2
    kpad_len2 = max(swz // 2 - (src_len - tgt_len), 0)
    attn_weights_pad = F.pad(attn_weights, (kpad_len1, kpad_len2), value=float("-inf"))
    attn_weights_slide = torch.as_strided(
        attn_weights_pad,
        size=(bsz, tgt_len, swz + 1),
        stride=(tgt_len * attn_weights_pad.size(2), attn_weights_pad.size(2) + 1, 1),
    ).contiguous()
    attn_probs_slide = F.softmax(attn_weights_slide, dim=-1)

    return attn, attn_weights_slide, attn_weights


def slide_window_attention_forward(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    swz: int,
    attn_mask=None,
    key_padding_mask=None,
    num_heads=None,
    dropout_p=0,
    training=False,
    mode="stride",
):
    r"""
    Sliding window attention.

    Args:
        q, k, v:
            3-dim tensor, e.g., (batch_size * head_num, seq_len, head_dim)
            q is a scaled q by sqrt(head_dim).
        swz:
            int, swz % 2 == 0, sliding window size
        attn_mask:
            2-d bool tensor (tgt_len, src_len) or
            3-d bool tensor (bsz, tgt_len, src_len)
        key_padding_mask:
            1-d long tensor (bsz) or
            2-d bool tensor (bsz, src_len)
        num_heads:
            int, required to merge `attn_mask` and `key_padding_mask`
        dropout_p:
            float, [0, 1], drop out `attn_probs` randomly.
        training:
            bool, it serves in `dropout`
        mode:
            'mask': via `attn_mask`
            'stride': via `torch.as_strided`
    Notes:
        1. `torch.as_strided` does not copy data, like `torch.Tensor.view`, and
        requires padding utterances to be the length of the multiple of `swz`.
        2. Pad K matrix with value of -inf, but pad V matrix with value of 0.
        3. Numerical error occurs while small `swz` is applied with `float16`,
        e.g., swz = 16, seq_len = 512, q.dtype = torch.float16
    """
    assert mode in ["mask", "stride"]
    assert swz > 0 and swz % 2 == 0, f"{swz} ? > 0 and % 2 == 0"
    q, k, v, attn_mask, key_padding_mask, ori_tgt_len, ori_src_len = pad_as_attn_swz(
        q, k, v, swz, attn_mask=attn_mask, key_padding_mask=key_padding_mask
    )
    attn_mask = merge_padding_attm_mask(
        attn_mask, key_padding_mask, num_heads, q.size(1)
    )
    _, tgt_len, src_len, _ = size_check(q, k, v)
    if mode == "mask":
        attn, attn_weights = sliding_attn_check_mask(
            q,
            k,
            v,
            swz,
            attn_mask=attn_mask,
            dropout_p=dropout_p,
            training=training,
            ret_slide=False,
        )
        attn = attn[:, tgt_len - ori_tgt_len :].contiguous()
        attn_weights = attn_weights[
            :, tgt_len - ori_tgt_len :, src_len - ori_src_len :
        ].contiguous()
    else:  # mode == "stride"
        assert (
            tgt_len == src_len and src_len % swz == 0
        ), f"{tgt_len} == {src_len} and {src_len} % {swz} == 0"
        attn, attn_weights = sliding_attn(
            q, k, v, swz, attn_mask=attn_mask, dropout_p=dropout_p, training=training
        )
        attn = attn[:, tgt_len - ori_tgt_len :].contiguous()
        attn_weights = attn_weights[:, tgt_len - ori_tgt_len :].contiguous()
    return attn, attn_weights


def global_attention_forward(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    attn_mask=None,
    key_padding_mask=None,
    num_heads=None,
    dropout_p=0,
    training=False,
):
    r"""
    Full/Global attention dot product as sliding attention with
        full-utterance window size.

    Args:
        q, k, v: tensor, (batch_size * head_num, seq_len, head_dim)
        attn_mask:
            2-d bool tensor (tgt_len, src_len) or
            3-d bool tensor (bsz, tgt_len, src_len)
        key_padding_mask:
            2-d bool tensor (bsz, src_len)

    Returns:
        attn, attn_weights: attention and unnormalized attention weights.
    """
    attn_weights = torch.bmm(q, k.transpose(1, 2))
    attn_mask = merge_padding_attm_mask(
        attn_mask, key_padding_mask, num_heads, q.size(1)
    )
    if attn_mask is not None:
        if attn_mask.ndim == 2:
            attn_mask = attn_mask.unsqueeze(0)
        attn_mask = attn_mask.to(attn_weights.dtype).masked_fill(
            attn_mask, float("-inf")
        )
        attn_weights += attn_mask
    attn_probs = F.softmax(attn_weights.to(torch.float32), dim=-1)
    attn_probs = attn_probs.to(attn_weights.dtype)
    attn_probs = F.dropout(attn_probs, p=0, training=False)
    attn = torch.bmm(attn_probs, v)

    return attn, attn_weights


def latency(
    func, inputs: dict, device="cpu", warmup_steps=50, measure_steps=100, backward=False
):
    import time

    if not backward:
        with torch.no_grad():
            if device != "cpu":
                torch.cuda.synchronize(device=device)
            for i in range(warmup_steps):
                func(**inputs)
            if device != "cpu":
                torch.cuda.synchronize(device=device)

            if device != "cpu":
                torch.cuda.synchronize(device=device)
            st = time.time()
            for i in range(measure_steps):
                func(**inputs)
            if device != "cpu":
                torch.cuda.synchronize(device=device)
            ed = time.time()
            total_time = ed - st
    else:
        original_grad = list(inputs.values())[0].requires_grad
        for key, value in inputs.items():
            if isinstance(value, torch.Tensor) and value.dtype != torch.bool:
                value.requires_grad = True
        if device != "cpu":
            torch.cuda.synchronize(device=device)
        for i in range(warmup_steps):
            ans = func(**inputs)[0].mean()
            ans.backward()
        if device != "cpu":
            torch.cuda.synchronize(device=device)

        if device != "cpu":
            torch.cuda.synchronize(device=device)
        st = time.time()
        for i in range(measure_steps):
            ans = func(**inputs)[0].mean()
            ans.backward()
        if device != "cpu":
            torch.cuda.synchronize(device=device)
        ed = time.time()
        total_time = ed - st
        for key, value in inputs.items():
            if isinstance(value, torch.Tensor) and value.dtype != torch.bool:
                value.requires_grad = original_grad

    avg_time = total_time / measure_steps * 1000  # ms

    return avg_time


if __name__ == "__main__":
    """
    python fairseq/models/compress_hubert/functional/sliding_attn.py --device cuda --swz 64 --len 510 --bsz 48 --half
    """
    print(__doc__)
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--profile", action="store_true")
    parser.add_argument(
        "--cross", action="store_true", help="cross attention uses different q, k, v"
    )
    parser.add_argument("--half", action="store_true")
    parser.add_argument("--double", action="store_true")
    parser.add_argument("--device", default="cuda")
    parser.add_argument("--swz", default=64, type=int)
    parser.add_argument("--len", default=510, type=int)
    parser.add_argument(
        "--ken", default=None, type=int, help="applied while cross attention"
    )
    parser.add_argument("--bsz", default=48, type=int)
    args = parser.parse_args()

    size = (args.bsz, args.len, 64)
    dsize = (args.bsz, args.len if args.ken is None else args.ken, 64)
    if args.cross:
        q, k, v = torch.rand(size), torch.rand(dsize), torch.rand(dsize)
    else:
        q = k = v = torch.rand(size)
    device = args.device
    q, k, v = q.to(device), k.to(device), v.to(device)
    if args.double:
        q, k, v = q.double(), k.double(), v.double()
    if args.half:
        q, k, v = q.half(), k.half(), v.half()
    swz = args.swz

    attn_mask = torch.rand((q.size(1), k.size(1))).to(device=device).to(q.dtype) > 0.5

    avg_time = latency(
        global_attention_forward,
        {"q": q, "k": k, "v": v, "attn_mask": attn_mask},
        device=device,
    )
    print(f"global_attention_forward: {avg_time} ms")
    avg_time = latency(
        slide_window_attention_forward,
        {"q": q, "k": k, "v": v, "swz": swz, "attn_mask": attn_mask, "mode": "stride"},
        device=device,
    )
    print(f"sliding_attn [slide]: {avg_time} ms")
    avg_time = latency(
        slide_window_attention_forward,
        {"q": q, "k": k, "v": v, "swz": swz, "attn_mask": attn_mask, "mode": "mask"},
        device=device,
    )
    print(f"sliding_attn [mask]: {avg_time} ms")
    avg_time = latency(
        sliding_attn_check_mask,
        {"q": q, "k": k, "v": v, "swz": swz, "attn_mask": attn_mask, "ret_slide": True},
        device=device,
    )
    print(f"sliding_attn_check_mask: {avg_time} ms")

    a0, aw0 = slide_window_attention_forward(
        q, k, v, swz, attn_mask=attn_mask, mode="stride"
    )
    a1, aw1 = slide_window_attention_forward(
        q, k, v, swz, attn_mask=attn_mask, mode="mask"
    )
    a2, aw2, aw_full = sliding_attn_check_mask(q, k, v, swz, attn_mask=attn_mask)

    print("Sliding vs. Masked:", torch.allclose(a0, a1))
    diff_a = "%.3f - %.3f = %.6e" % (
        a0.view(-1)[(a0 - a2).abs().argmax()].item(),
        a2.view(-1)[(a0 - a2).abs().argmax()].item(),
        (a0 - a2).abs().max().item(),
    )
    aw0s = aw0.softmax(-1)
    aw2s = aw2.softmax(-1)
    diff_aw = "%.3f - %.3f = %.6e" % (
        aw0s.view(-1)[(aw0s - aw2s).abs().argmax()].item(),
        aw2s.view(-1)[(aw0s - aw2s).abs().argmax()].item(),
        (aw0s - aw2s).abs().max().item(),
    )
    print(
        f"Sliding vs. Masked: {torch.allclose(a0, a2)} ({diff_a}) {torch.allclose(aw0s, aw2s)} ({diff_aw})"
    )

    if args.profile:
        with torch.autograd.profiler.profile(
            use_cuda=True, with_stack=False, profile_memory=True
        ) as prof:
            a0, aw0 = slide_window_attention_forward(
                q, k, v, swz, attn_mask=attn_mask, mode="stride"
            )
        print("sliding_attn", prof.table())

        with torch.autograd.profiler.profile(
            use_cuda=True, with_stack=False, profile_memory=True
        ) as prof:
            a2, aw2 = slide_window_attention_forward(
                q, k, v, swz, attn_mask=attn_mask, mode="mask"
            )
        print("sliding_attn_check_mask", prof.table())



"""
Common probing models

Authors:
  * Leo 2022
"""

from typing import List

import torch.nn as nn

import s3prl.nn.pooling as pooling

__all__ = [
    "FrameLevel",
    "UtteranceLevel",
]


class FrameLevel(nn.Module):
    """
    The common frame-to-frame probing model

    Args:
        input_size (int): input size
        output_size (int): output size
        hidden_sizes (List[int]): a list of hidden layers' hidden size.
            by default is [256] to project all different input sizes to the same dimension.
            set empty list to use the vanilla single layer linear model
        activation_type (str): the activation class name in :obj:`torch.nn`. Set None to
            disable activation and the model is pure linear. Default: None
        activation_conf (dict): the arguments for initializing the activation class.
            Default: empty dict
    """

    def __init__(
        self,
        input_size: int,
        output_size: int,
        hidden_sizes: List[int] = None,
        activation_type: str = None,
        activation_conf: dict = None,
    ):
        super().__init__()
        self._indim = input_size
        self._outdim = output_size
        hidden_sizes = hidden_sizes or [256]

        latest_size = input_size
        hidden_layers = []
        if len(hidden_sizes) > 0:
            for size in hidden_sizes:
                hidden_layers.append(nn.Linear(latest_size, size))
                if activation_type is not None:
                    hidden_layers.append(
                        getattr(nn, activation_type)(**(activation_conf or {}))
                    )
                latest_size = size

        self.hidden_layers = nn.Sequential(*hidden_layers)
        self.final_proj = nn.Linear(latest_size, output_size)

    @property
    def input_size(self) -> int:
        return self._indim

    @property
    def output_size(self) -> int:
        return self._outdim

    def forward(self, x, x_len):
        """
        Args:
            x (torch.FloatTensor): (batch_size, seq_len, input_size)
            x_len (torch.LongTensor): (batch_size, )

        Returns:
            tuple

            1. ys (torch.FloatTensor): (batch_size, seq_len, output_size)
            2. ys_len (torch.LongTensor): (batch_size, )
        """
        ys = self.hidden_layers(x)
        ys = self.final_proj(ys)
        return ys, x_len


class UtteranceLevel(nn.Module):
    """
    Args:
        input_size (int): input_size
        output_size (int): output_size
        hidden_sizes (List[int]): a list of hidden layers' hidden size.
            by default is [256] to project all different input sizes to the same dimension.
            set empty list to use the vanilla single layer linear model
        activation_type (str): the activation class name in :obj:`torch.nn`. Set None to
            disable activation and the model is pure linear. Default: None
        activation_conf (dict): the arguments for initializing the activation class.
            Default: empty dict
        pooling_type (str): the pooling class name in :obj:`s3prl.nn.pooling`. Default: MeanPooling
        pooling_conf (dict): the arguments for initializing the pooling class.
            Default: empty dict
    """

    def __init__(
        self,
        input_size: int,
        output_size: int,
        hidden_sizes: List[int] = None,
        activation_type: str = None,
        activation_conf: dict = None,
        pooling_type: str = "MeanPooling",
        pooling_conf: dict = None,
    ):
        super().__init__()
        self._indim = input_size
        self._outdim = output_size
        hidden_sizes = hidden_sizes or [256]

        latest_size = input_size
        hidden_layers = []
        if len(hidden_sizes) > 0:
            for size in hidden_sizes:
                hidden_layers.append(nn.Linear(latest_size, size))
                if activation_type is not None:
                    hidden_layers.append(
                        getattr(nn, activation_type)(**(activation_conf or {}))
                    )
                latest_size = size

        self.hidden_layers = nn.Sequential(*hidden_layers)

        pooling_conf = pooling_conf or {}
        self.pooling = getattr(pooling, pooling_type)(latest_size, **pooling_conf)
        latest_size = self.pooling.output_size
        self.final_proj = nn.Linear(latest_size, output_size)

    @property
    def input_size(self) -> int:
        return self._indim

    @property
    def output_size(self) -> int:
        return self._outdim

    def forward(self, x, x_len):
        """
        Args:
            x (torch.FloatTensor): (batch_size, seq_len, input_size)
            x_len (torch.LongTensor): (batch_size, )

        Returns:
            torch.FloatTensor

            (batch_size, output_size)
        """
        x = self.hidden_layers(x)
        x_pooled = self.pooling(x, x_len)
        y = self.final_proj(x_pooled)
        return y



import logging
from collections import OrderedDict
from copy import deepcopy
from functools import partial
from typing import Any, List, Union

import numpy as np
import torch
from speechbrain.dataio.dataset import DynamicItemDataset
from speechbrain.utils.data_pipeline import DynamicItem
from torch.nn.utils.rnn import pad_sequence

logger = logging.getLogger(__name__)


class AugmentedDynamicItemDataset(DynamicItemDataset):
    def __init__(
        self,
        data,
        dynamic_items=[],
        output_keys=[],
        tools: dict = {},
    ):
        super().__init__(data, dynamic_items, output_keys)
        assert isinstance(data, OrderedDict)
        self._tools = {}
        for name, item in tools.items():
            self.add_tool(name, item)

    def _dynamic_tools(self, id, name):
        return self._tools[name]

    def add_tool(self, name: str, item: Any) -> None:
        """
        Store the :code:`item` in this dataset with the name :code:`name` so it can be used in
        :code:`__getitem__`. That is, you can retrieve the :code:`item` with the :code:`takes` argument
        of :obj:`add_dynamic_item`.

        .. code-block:: python

            def tokenize_func(tokenizer, text):
                return tokenizer(text)

            self.add_tool("tokenizer", tokenizer)
            self.add_dynamic_item(tokenize_func, takes=["tokenizer", "text"], provides="tokenized_ids")

        You can also later retreive this tool by :obj:`get_tool` or :obj:`all_tools`
        """
        self._tools[name] = item
        self.add_dynamic_item(
            partial(self._dynamic_tools, name=name), takes="id", provides=name
        )

    def add_tools(self, tools: dict) -> None:
        """
        Store each key-value pair in :code:`tools` as a tool. See :obj:`add_tool` for more information
        """
        for key, value in tools.items():
            self.add_tool(key, value)

    def get_tool(self, key) -> Any:
        """
        See :obj:`add_tool` for more information
        """
        return self._tools[key]

    def has_tool(self, key) -> bool:
        """
        Checks whether has a tool named :code:`key`.
        """
        return key in self._tools

    def all_tools(self, copy=True) -> dict:
        """
        Return:
            dict

            Containing all the tools in :code:`name: value` pairs.
            See :obj:`add_tool` for more information
        """
        return deepcopy(self._tools) if copy else self._tools

    def update_output_keys(self, keys: dict) -> None:
        """
        Compared to :obj:`set_output_keys`, this method update the output keys mapping
        instead of replace it with a new dictionary. This can be useful when you only
        want to replace a few mapping and leave others unchanged.
        """
        mapping = self.pipeline.output_mapping.copy()
        mapping.update(keys)
        self.set_output_keys(mapping)

    def keys(self) -> List[str]:
        """
        List all the :code:`static_item` and :code:`dynamic_item` in the dataset.
        :code:`static_item` resides directly in the memory and are given by the dataset
        initialization dictionary. :code:`dynamic_item` are content computed
        on-the-fly basing on :code:`static_item`.
        """
        available_keys: List[str] = list(self.pipeline.key_to_node.keys())
        for dynamic_item in self.pipeline.dynamic_items:
            provides = dynamic_item.provides
            assert isinstance(provides, (list, tuple))
            available_keys += provides
        available_keys = [
            key
            for key in available_keys
            if not key.startswith("_") and key not in self._tools
        ]
        return available_keys

    def set_info(self, info):
        self._info = info

    def get_info(self, index):
        with self.output_keys_as(self._info):
            return self.__getitem__(index)

    def __getitem__(self, index):
        """
        This remain all the usage of the original SpeechBrain DynamicItemDataset.__getitem__,
        except that by default it uses :obj:`keys` as the default :code:`output_keys`
        """
        if len(self.pipeline.output_mapping) == 0:
            with self.output_keys_as(self.keys()):
                return super().__getitem__(index)
        else:
            return super().__getitem__(index)


class DataPipe:
    def __call__(
        self, dataset: Union[dict, AugmentedDynamicItemDataset], tools: dict = None
    ) -> Any:
        if isinstance(dataset, dict):
            dataset = AugmentedDynamicItemDataset(dataset)

        if tools is not None:
            dataset.add_tools(tools)

        return self.forward(dataset)

    def forward(
        self, dataset: AugmentedDynamicItemDataset
    ) -> AugmentedDynamicItemDataset:
        raise NotImplementedError

    def __getattribute__(self, name):
        value = super().__getattribute__(name)
        if isinstance(value, DynamicItem):
            value.func = value.func.__get__(self)
        return value


class SequentialDataPipe(DataPipe):
    def __init__(self, *pipes: List[DataPipe]) -> None:
        self._pipes = pipes

    def forward(
        self, dataset: AugmentedDynamicItemDataset
    ) -> AugmentedDynamicItemDataset:
        for pipe in self._pipes:
            dataset = pipe(dataset)
        return dataset


def default_collate_fn(samples, padding_value: int = 0):
    """
    Each item in **DynamicItemDataset** is a dict
    This function pad (or transform into numpy list) a batch of dict

    Args:
        samples (List[dict]): Suppose each Container is in

            .. code-block:: yaml

                wav: a single waveform
                label: a single string

    Return:
        dict

        .. code-block:: yaml

            wav: padded waveforms
            label: np.array([a list of string labels])
    """
    assert isinstance(samples[0], dict)
    keys = samples[0].keys()
    padded_samples = dict()
    for key in keys:
        values = [sample[key] for sample in samples]
        if isinstance(values[0], int):
            values = torch.LongTensor(values)
        elif isinstance(values[0], float):
            values = torch.FloatTensor(values)
        elif isinstance(values[0], np.ndarray):
            values = [torch.from_numpy(value).float() for value in values]
            values = pad_sequence(values, batch_first=True, padding_value=padding_value)
        elif isinstance(values[0], torch.Tensor):
            values = pad_sequence(values, batch_first=True, padding_value=padding_value)
        else:
            values = np.array(values, dtype="object")
        padded_samples[key] = values
    return padded_samples



# --------------------------------------------------------
# LightHuBERT: Lightweight and Configurable Speech Representation Learning with Once-for-All Hidden-Unit BERT (https://arxiv.org/pdf/2203.15610.pdf)
# Github source: https://github.com/mechanicalsea/lighthubert
# Licensed under The MIT License [see LICENSE for details]
# Based on fairseq code bases
# https://github.com/pytorch/fairseq
# --------------------------------------------------------

from typing import List, Tuple

import torch.nn as nn

from .fairseq_modules import Fp32GroupNorm, Fp32LayerNorm, TransposeLast


class ConvFeatureExtractionModel(nn.Module):
    def __init__(
        self,
        conv_layers: List[Tuple[int, int, int]],
        dropout: float = 0.0,
        mode: str = "default",
        conv_bias: bool = False,
    ):
        super().__init__()

        assert mode in {"default", "layer_norm"}

        def block(
            n_in,
            n_out,
            k,
            stride,
            is_layer_norm=False,
            is_group_norm=False,
            conv_bias=False,
        ):
            def make_conv():
                conv = nn.Conv1d(n_in, n_out, k, stride=stride, bias=conv_bias)
                nn.init.kaiming_normal_(conv.weight)
                return conv

            assert (
                is_layer_norm and is_group_norm
            ) == False, "layer norm and group norm are exclusive"

            if is_layer_norm:
                return nn.Sequential(
                    make_conv(),
                    nn.Dropout(p=dropout),
                    nn.Sequential(
                        TransposeLast(),
                        Fp32LayerNorm(dim, elementwise_affine=True),
                        TransposeLast(),
                    ),
                    nn.GELU(),
                )
            elif is_group_norm:
                return nn.Sequential(
                    make_conv(),
                    nn.Dropout(p=dropout),
                    Fp32GroupNorm(dim, dim, affine=True),
                    nn.GELU(),
                )
            else:
                return nn.Sequential(make_conv(), nn.Dropout(p=dropout), nn.GELU())

        in_d = 1
        self.conv_layers = nn.ModuleList()
        for i, cl in enumerate(conv_layers):
            assert len(cl) == 3, "invalid conv definition: " + str(cl)
            (dim, k, stride) = cl

            self.conv_layers.append(
                block(
                    in_d,
                    dim,
                    k,
                    stride,
                    is_layer_norm=mode == "layer_norm",
                    is_group_norm=mode == "default" and i == 0,
                    conv_bias=conv_bias,
                )
            )
            in_d = dim

    def forward(self, x):
        # BxT -> BxCxT
        x = x.unsqueeze(1)

        for conv in self.conv_layers:
            x = conv(x)

        return x



import tempfile

import torch
import torchaudio

SAMPLE_RATE = 16000


def test_sox_effect():
    effects = [
        ["channels", "1"],
        ["rate", "16000"],
        ["gain", "-3.0"],
    ]

    with tempfile.NamedTemporaryFile() as file:
        tensor = torch.randn(1, 16000 * 10)
        filename = f"{file.name}.wav"
        torchaudio.save(filename, tensor, SAMPLE_RATE)
        wav1, sr1 = torchaudio.sox_effects.apply_effects_file(filename, effects=effects)
        wav2, sr2 = torchaudio.sox_effects.apply_effects_tensor(
            tensor, SAMPLE_RATE, effects
        )
        torch.allclose(wav1, wav2)
        assert sr1 == sr2



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ upstream/baseline/preprocessor.py ]
#   Synopsis     [ the baseline feature preprocessor with torchaudio.transforms as backend ]
#   Author       [ S3PRL ]
#   Copyright    [ Copyleft(c), Speech Lab, NTU, Taiwan ]
"""*********************************************************************************************"""


from functools import partial

###############
# IMPORTATION #
###############
import torch
import torchaudio
from torch.nn.utils.rnn import pad_sequence
from torchaudio.functional import compute_deltas
from torchaudio.transforms import MFCC, MelScale, Spectrogram

############
# CONSTANT #
############
N_SAMPLED_PSEUDO_WAV = 2


def get_preprocessor(audio_config, take_first_channel=True, process_input_only=False):
    """
    Args:
        take_first_channel:
            bool
            If True, the preprocessor takes input as: (*, channel=1, waveform_len),
                where `input` and `target` are taken from the same channel (i.e. the same and single waveform).
            If False, the preprocessor takes input as: (*, channel=2, waveform_len),
                where `input` and `target` are taken from the specified channel (i.e. multiple correlated views of the same waveform).
                For example: preprocessor_input = torch.cat([wav_aug1.unsqueeze(1), wav_aug2.unsqueeze(1)], dim=1) # shape: (batch_size, channel=2, seq_len)
        process_input_only:
            bool
            If True, the preprocessor will process `input`
            If False, the preprocessor will process both `input` and `target`
    """
    assert not (take_first_channel == False and process_input_only == True)

    if not "target" in audio_config:
        audio_config["target"] = audio_config["input"]

    input_feat = audio_config["input"]
    target_feat = audio_config["target"]
    if take_first_channel:
        input_feat["channel"] = 0
        target_feat["channel"] = 0

    preprocessor = OnlinePreprocessor(
        **audio_config, feat_list=[input_feat, target_feat]
    )
    input_dim, target_dim = [feat.size(-1) for feat in preprocessor()]

    if process_input_only:
        del preprocessor
        preprocessor = OnlinePreprocessor(**audio_config, feat_list=[input_feat])
    return preprocessor, input_dim, target_dim


class OnlinePreprocessor(torch.nn.Module):
    def __init__(
        self,
        sample_rate=16000,
        win_ms=25,
        hop_ms=10,
        n_freq=201,
        n_mels=40,
        n_mfcc=13,
        feat_list=None,
        eps=1e-10,
        **kwargs
    ):
        super(OnlinePreprocessor, self).__init__()
        # save preprocessing arguments
        self._sample_rate = sample_rate
        self._win_ms = win_ms
        self._hop_ms = hop_ms
        self._n_freq = n_freq
        self._n_mels = n_mels
        self._n_mfcc = n_mfcc

        win = round(win_ms * sample_rate / 1000)
        hop = round(hop_ms * sample_rate / 1000)
        n_fft = (n_freq - 1) * 2
        self._win_args = {"n_fft": n_fft, "hop_length": hop, "win_length": win}
        self.register_buffer("_window", torch.hann_window(win))

        self._stft_args = {
            "center": True,
            "pad_mode": "reflect",
            "normalized": False,
            "onesided": True,
            "return_complex": True,
        }
        # stft_args: same default values as torchaudio.transforms.Spectrogram & librosa.core.spectrum._spectrogram
        self._stft = partial(torch.stft, **self._win_args, **self._stft_args)
        self._melscale = MelScale(sample_rate=sample_rate, n_mels=n_mels)
        self._mfcc_trans = MFCC(
            sample_rate=sample_rate,
            n_mfcc=n_mfcc,
            log_mels=True,
            melkwargs=self._win_args,
        )
        self._istft = partial(torch.istft, **self._win_args, **self._stft_args)

        self.feat_list = feat_list
        self.register_buffer(
            "_pseudo_wavs", torch.randn(N_SAMPLED_PSEUDO_WAV, sample_rate)
        )
        self.eps = eps

    def _magphase(self, spectrogram):
        # Replace the deprecated private member function self._magphase
        # Ref. https://github.com/pytorch/audio/issues/1337
        # original: `self._magphase = partial(torchaudio.functional.magphase, power=2)`
        spectrogram = torch.view_as_complex(spectrogram)
        return spectrogram.abs().pow(exponent=2), spectrogram.angle()

    def _check_list(self, feat_list):
        if feat_list is None:
            feat_list = self.feat_list
        assert type(feat_list) is list
        return feat_list

    def _transpose_list(self, feats):
        return [
            feat.transpose(-1, -2).contiguous() if type(feat) is torch.Tensor else feat
            for feat in feats
        ]

    @classmethod
    def get_feat_config(cls, feat_type, channel=0, log=False, delta=0, cmvn=False):
        assert feat_type in ["wav", "complx", "linear", "phase", "mel", "mfcc"]
        assert type(channel) is int
        assert type(log) is bool
        assert type(delta) is int and delta >= 0
        assert type(cmvn) is bool
        return {
            "feat_type": feat_type,
            "channel": channel,
            "log": log,
            "delta": delta,
            "cmvn": cmvn,
        }

    def forward(self, wavs=None, feat_list=None, wavs_len=None):
        # wavs: (*, channel_size, max_len)
        # feat_list, mam_list: [{feat_type: 'mfcc', channel: 0, log: False, delta: 2, cmvn: 'True'}, ...]
        # wavs_len: [len1, len2, ...]

        feat_list = self._check_list(feat_list)
        if wavs is None:
            max_channel_id = max(
                [int(args["channel"]) if "channel" in args else 0 for args in feat_list]
            )
            wavs = self._pseudo_wavs[0].view(1, 1, -1).repeat(1, max_channel_id + 1, 1)
        assert wavs.dim() >= 3

        # find length of wavs from padded tensor
        if wavs_len is None:
            wavs_len = []
            for wav in wavs:
                nonzero_index = wav.nonzero()
                if len(nonzero_index) == 0:
                    wavs_len.append(wav.size(-1))  # when all elements are zero
                else:
                    wavs_len.append(nonzero_index[:, -1].max().item() + 1)
            wavs = pad_sequence(
                [
                    wav[:wav_len].transpose(-1, -2)
                    for wav, wav_len in zip(wavs, wavs_len)
                ],
                batch_first=True,
            ).transpose(-1, -2)

        wav = wavs.unsqueeze(2)
        shape = wavs.size()
        complx = self._stft(wavs.reshape(-1, shape[-1]), window=self._window)
        complx = torch.view_as_real(complx)
        complx = complx.reshape(shape[:-1] + complx.shape[-3:])
        # complx: (*, channel_size, feat_dim, max_len, 2)
        linear, phase = self._magphase(complx)
        mel = self._melscale(linear)
        mfcc = self._mfcc_trans(wavs)
        complx = complx.transpose(-1, -2).reshape(*mfcc.shape[:2], -1, mfcc.size(-1))
        # complx, linear, phase, mel, mfcc: (*, channel_size, feat_dim, max_len)

        def select_feat(
            variables, feat_type, channel=0, log=False, delta=0, cmvn=False
        ):
            raw_feat = variables[feat_type].select(dim=-3, index=channel)
            # apply log scale
            if bool(log):
                raw_feat = (raw_feat + self.eps).log()
            feats = [raw_feat.contiguous()]
            # apply delta for features
            for _ in range(int(delta)):
                feats.append(compute_deltas(feats[-1]))
            feats = torch.cat(feats, dim=-2)
            downsample_rate = wavs.size(-1) / feats.size(-1)
            feats_len = [round(length / downsample_rate) for length in wavs_len]
            # apply cmvn
            if bool(cmvn):
                cmvn_feats = []
                for feat, feat_len in zip(feats, feats_len):
                    feat = feat[:, :feat_len]
                    cmvn_feat = (feat - feat.mean(dim=-1, keepdim=True)) / (
                        feat.std(dim=-1, keepdim=True) + self.eps
                    )
                    cmvn_feats.append(cmvn_feat.transpose(-1, -2))
                feats = pad_sequence(cmvn_feats, batch_first=True).transpose(-1, -2)
            return feats
            # return: (*, feat_dim, max_len)

        local_variables = locals()
        return self._transpose_list(
            [select_feat(local_variables, **args) for args in feat_list]
        )
        # return: [(*, max_len, feat_dim), ...]

    def istft(self, linears=None, phases=None, linear_power=2, complxs=None):
        assert complxs is not None or (linears is not None and phases is not None)
        # complxs: (*, n_freq, max_feat_len, 2) or (*, max_feat_len, n_freq * 2)
        # linears, phases: (*, max_feat_len, n_freq)

        if complxs is None:
            linears, phases = self._transpose_list([linears, phases])
            complxs = linears.pow(1 / linear_power).unsqueeze(-1) * torch.stack(
                [phases.cos(), phases.sin()], dim=-1
            )
        if complxs.size(-1) != 2:
            # treat complxs as: (*, max_feat_len, n_freq * 2)
            shape = complxs.size()
            complxs = complxs.view(*shape[:-1], -1, 2).transpose(-2, -3).contiguous()
        # complxs: (*, n_freq, max_feat_len, 2)

        return self._istft(complxs, window=self._window)
        # return: (*, max_wav_len)

    def test_istft(self, wavs=None, atol=1e-6):
        # wavs: (*, channel_size, max_wav_len)
        channel1, channel2 = 0, 1
        max_channel_id = max(channel1, channel2)

        if wavs is None:
            wavs = self._pseudo_wavs[: max_channel_id + 1].unsqueeze(0)
        assert wavs.size(-2) > max_channel_id

        feat_list = [
            {"feat_type": "complx", "channel": channel1},
            {"feat_type": "linear", "channel": channel2},
            {"feat_type": "phase", "channel": channel2},
        ]
        complxs, linears, phases = self.forward(wavs, feat_list)
        assert torch.allclose(
            wavs.select(dim=-2, index=channel1), self.istft(complxs=complxs), atol=atol
        )
        assert torch.allclose(
            wavs.select(dim=-2, index=channel2),
            self.istft(linears=linears, phases=phases),
            atol=atol,
        )
        print("[Preprocessor] test passed: stft -> istft")



"""
Create pseudo data

Authors
  * Leo 2022
"""

import random
import shutil
import tempfile
from pathlib import Path
from typing import Any, List

import torch
import torchaudio
from torch.nn.utils.rnn import pad_sequence

SAMPLE_RATE = 16000

__all__ = [
    "pseudo_audio",
    "get_pseudo_wavs",
]


class pseudo_audio:
    """
    This context manager returns filepaths (List[str]) and num_samples (List[int]) on entering
    """

    def __init__(self, secs: List[float], sample_rate: int = SAMPLE_RATE):
        self.tempdir = Path(tempfile.TemporaryDirectory().name)
        self.tempdir.mkdir(parents=True, exist_ok=True)
        self.num_samples = []
        for n, sec in enumerate(secs):
            wav = torch.randn(1, round(sample_rate * sec))
            torchaudio.save(
                str(self.tempdir / f"audio_{n}.wav"), wav, sample_rate=sample_rate
            )
            self.num_samples.append(wav.size(-1))
        self.filepaths = [
            str(self.tempdir / f"audio_{i}.wav") for i in range(len(secs))
        ]

    def __enter__(self):
        return self.filepaths, self.num_samples

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        shutil.rmtree(self.tempdir)


def get_pseudo_wavs(
    seed: int = 0,
    n: int = 2,
    min_secs: int = 1,
    max_secs: int = 3,
    sample_rate: int = SAMPLE_RATE,
    device: str = "cpu",
    padded: bool = False,
):
    random.seed(seed)
    torch.manual_seed(seed)

    wavs = []
    wavs_len = []
    for _ in range(n):
        wav_length = random.randint(
            round(min_secs * sample_rate), round(max_secs * sample_rate)
        )
        wav = torch.randn(wav_length, requires_grad=True).to(device)
        wavs_len.append(wav_length)
        wavs.append(wav)

    if not padded:
        return wavs
    else:
        return pad_sequence(wavs, batch_first=True), torch.LongTensor(wavs_len)



"""
Parse the commonly used corpus into standardized dictionary structure
"""

from .fluent_speech_commands import FluentSpeechCommands
from .iemocap import IEMOCAP
from .librilight import LibriLight
from .librispeech import LibriSpeech
from .quesst14 import Quesst14
from .snips import SNIPS
from .speech_commands import SpeechCommandsV1
from .voxceleb1sid import VoxCeleb1SID
from .voxceleb1sv import VoxCeleb1SV

__all__ = [
    "FluentSpeechCommands",
    "IEMOCAP",
    "LibriSpeech",
    "LibriLight",
    "Quesst14",
    "SNIPS",
    "SpeechCommandsV1",
    "VoxCeleb1SID",
    "VoxCeleb1SV",
]



import torch

from s3prl.nn import RNNEncoder


def test_rnn(helpers):
    modules = [
        RNNEncoder(
            input_size=8,
            output_size=6,
            module="LSTM",
            hidden_size=[10, 10, 10],
            dropout=[0.1, 0.1, 0.1],
            layer_norm=[True, True, True],
            proj=[True, True, True],
            sample_rate=[1, 2, 1],
            sample_style="drop",
            bidirectional=True,
        ),
        RNNEncoder(
            input_size=8,
            output_size=6,
            module="LSTM",
            hidden_size=[10, 10, 10],
            dropout=[0.1, 0.1, 0.1],
            layer_norm=[True, True, True],
            proj=[True, True, True],
            sample_rate=[1, 2, 1],
            sample_style="concat",
            bidirectional=True,
        ),
    ]

    for module in modules:
        xs = torch.randn(32, 50, module.input_size)
        xs_len = torch.arange(32) + (50 - 32) + 1

        out, out_len = module(xs, xs_len)
        assert out.shape[1] == 25
        assert out.shape[2] == module.output_size
        assert out_len.max() == 25



# -*- coding: utf-8 -*-
# @Time    : 8/25/21 5:25 PM
# @Author  : Yuan Gong
# @Affiliation  : Massachusetts Institute of Technology
# @Email   : yuangong@mit.edu
# @File    : expert.py

# Authors
# - Leo

import logging

import torch

from .audio import FeatureExtractor

logger = logging.getLogger(__name__)

FBANK_SAMPLE_STRIDE = 160
SAMPLE_RATE = 16000


class UpstreamExpert(torch.nn.Module):
    def __init__(
        self,
        ckpt: str,
        model_size: str,
        window_secs: float = 1.0,
    ):
        super().__init__()
        self.window_secs = window_secs
        self.stride_secs = window_secs

        try:
            import timm

            from .ast_models import ASTModel
        except:
            logger.error(
                "SSAST requires 'timm==0.4.5' to work. Please run 'pip install timm==0.4.5'"
            )
            exit(1)

        target_length = int(window_secs * SAMPLE_RATE / FBANK_SAMPLE_STRIDE)
        model_size, model_type = model_size.split("_")[0], model_size.split("_")[1]
        self.preprocessor = FeatureExtractor(
            target_length=target_length, apply_cmvn=False
        )

        assert model_type in ["p", "f"]
        if model_type == "p":
            self.tstride = 10
            self.model = ASTModel(
                fshape=16,
                tshape=16,
                fstride=10,
                tstride=self.tstride,
                input_tdim=target_length,
                input_fdim=128,
                model_size=model_size,
                pretrain_stage=False,
                load_pretrained_mdl_path=ckpt,
            )
            self.vertical_num_patches = (128 - 16) // 10 + 1  # 12
        else:
            self.tstride = 1
            self.model = ASTModel(
                fshape=128,
                tshape=2,
                fstride=128,
                tstride=self.tstride,
                input_tdim=target_length,
                input_fdim=128,
                model_size=model_size,
                pretrain_stage=False,
                load_pretrained_mdl_path=ckpt,
            )
            self.vertical_num_patches = (128 - 128) // 128 + 1  # 1

        self.model = self.model.cpu()

    def get_downsample_rates(self, key: str = None) -> int:
        return int(FBANK_SAMPLE_STRIDE * self.tstride)

    def forward(self, wavs):
        wavs_len = [len(wav) for wav in wavs]
        max_wav_len = max(wavs_len)
        start_points = list(range(0, max_wav_len, int(self.stride_secs * SAMPLE_RATE)))
        padded_max_wav_len = start_points[-1] + int(self.window_secs * SAMPLE_RATE)
        padded_wavs = [
            torch.cat([wav, wav.new_zeros(padded_max_wav_len - len(wav))])
            for wav in wavs
        ]

        all_features = []
        for start in start_points:
            subwavs = [
                wav[start : start + int(self.window_secs * SAMPLE_RATE)]
                for wav in padded_wavs
            ]
            features = [self.preprocessor(wav.unsqueeze(0)) for wav in subwavs]
            features = torch.stack(
                features, dim=0
            )  # (batch_size, segment_seq_len, hidden_size)
            all_features.append(features)

        all_features = torch.stack(all_features, dim=0)
        num_segment, batch_size, segment_seq_len, hidden_size = all_features.shape

        flatten_features = all_features.reshape(-1, segment_seq_len, hidden_size)
        hidden_states, final_repr = self.model(flatten_features)

        reshaped_hidden_states = [
            (
                h.reshape(num_segment, batch_size, -1, h.size(-1))
                .transpose(
                    0, 1
                )  # (batch_size, num_segment, num_horizon_patch, num_vertical_patch * hidden_size)
                .flatten(
                    1, 2
                )  # (batch_size, num_segment * num_horizon_patch, num_vertical_patch * hidden_size)
                .float()
            )
            for h in hidden_states
        ]

        trimmed_hidden_states = []
        for h in reshaped_hidden_states:
            max_h_len = len(range(0, max_wav_len, self.get_downsample_rates()))
            h = h[:, :max_h_len, :]
            trimmed_hidden_states.append(h)

        return {"hidden_states": trimmed_hidden_states}
